from __future__ import annotations

from pathlib import Path

from configargparse import ArgumentParser
from prefect import flow, tags, unmapped
from prefect.futures import PrefectFuture

from flint.coadd.linmos import LinmosOptions, LinmosResult
from flint.configuration import (
    POLARISATION_MAPPING,
    get_options_from_strategy,
    load_and_copy_strategy,
)
from flint.exceptions import MSError
from flint.imager.wsclean import (
    ImageSet,
    WSCleanResult,
)
from flint.logging import logger
from flint.ms import find_mss
from flint.naming import (
    CASDANameComponents,
    ProcessedNameComponents,
    add_timestamp_to_path,
    extract_components_from_name,
    get_sbid_from_path,
)
from flint.options import (
    PolFieldOptions,
    add_options_to_parser,
    create_options_from_parser,
    dump_field_options_to_yaml,
)
from flint.prefect.clusters import get_dask_runner
from flint.prefect.common.imaging import (
    task_combine_images_to_cube,
    task_convolve_images,
    task_get_channel_images_from_paths,
    task_get_common_beam_from_image_set,
    task_linmos_images,
    task_merge_image_sets,
    task_preprocess_askap_ms,
    task_split_and_get_image_set,
    task_wsclean_imager,
)
from flint.prefect.common.utils import (
    task_create_field_summary,
    task_create_object,
    task_getattr,
    task_rename_linear_to_stokes,
)


@flow(name="Flint Polarisation Pipeline")
def process_science_fields_pol(
    flint_ms_directory: Path,
    pol_field_options: PolFieldOptions,
) -> None:
    strategy = load_and_copy_strategy(
        output_split_science_path=flint_ms_directory,
        imaging_strategy=pol_field_options.imaging_strategy,
    )

    logger.info(f"{pol_field_options=}")

    if strategy is None:
        logger.info("No strategy provided. Returning.")
        return

    # Get some placeholder names
    science_mss = list(
        find_mss(
            mss_parent_path=flint_ms_directory,
            expected_ms_count=pol_field_options.expected_ms,
            data_column=strategy["defaults"].get("data_column", "DATA"),
        )
    )
    # Check if MSs have been processed by Flint or have been provided by CASDA
    from_flint_list = [
        isinstance(extract_components_from_name(ms.path), ProcessedNameComponents)
        for ms in science_mss
    ]
    from_casda_list = [
        isinstance(extract_components_from_name(ms.path), CASDANameComponents)
        for ms in science_mss
    ]

    if not any(from_flint_list) and not any(from_casda_list):
        raise MSError("No valid MeasurementSets found! Data must be calibrated first.")

    if any(from_flint_list) and any(from_casda_list):
        raise MSError("Cannot mix Flint-processed and CASDA-provided MeasurementSets!")

    if any(from_casda_list):
        assert all(from_casda_list), (
            "Some MeasurementSets are from Flint, some are from CASDA"
        )
        logger.info("Data are from CASDA, need to apply FixMS")
        if pol_field_options.casa_container is None:
            msg = "We need to apply FixMS to CASDA-provided data, but no CASA container provided"
            raise MSError(msg)

        corrected_mss = []
        for ms in science_mss:
            corrected_ms = task_preprocess_askap_ms.submit(
                ms=ms,
                data_column=strategy["defaults"].get("data_column", "DATA"),
                skip_rotation=False,
                fix_stokes_factor=True,
                apply_ms_transform=True,
                casa_container=pol_field_options.casa_container,
                rename=True,
            )
            corrected_mss.append(corrected_ms)

        assert len(corrected_mss) == len(science_mss), (
            "Number of corrected MSs does not match number of input MSs"
        )
        science_mss = corrected_mss

    field_summary = task_create_field_summary.submit(
        mss=science_mss,
        holography_path=pol_field_options.holofile,
    )

    dump_field_options_to_yaml(
        output_path=add_timestamp_to_path(
            input_path=flint_ms_directory / "pol_field_options.yaml"
        ),
        field_options=pol_field_options,
    )

    logger.info(f"Found the following calibrated measurement sets: {science_mss}")

    if pol_field_options.wsclean_container is None:
        logger.info("No wsclean container provided. Returning. ")
        return

    polarisations: dict[str, str] = strategy.get("polarisation", {"total": {}})

    image_sets_dict: dict[str, PrefectFuture[ImageSet]] = {}
    image_sets_list: list[PrefectFuture[ImageSet]] = []
    for polarisation in polarisations.keys():
        _image_sets = []
        with tags(f"polarisation-{polarisation}"):
            for science_ms in science_mss:
                wsclean_result: PrefectFuture[WSCleanResult] = (
                    task_wsclean_imager.submit(
                        in_ms=science_ms,
                        wsclean_container=pol_field_options.wsclean_container,
                        make_cube_from_subbands=False,  # We will do this later
                        update_wsclean_options=unmapped(
                            get_options_from_strategy(
                                strategy=strategy,
                                operation="polarisation",
                                mode="wsclean",
                                polarisation=polarisation,
                            )
                        ),
                    )
                )
                _image_set: PrefectFuture[ImageSet] = task_getattr.submit(
                    wsclean_result, "image_set"
                )
                _image_sets.append(_image_set)
                image_sets_list.append(_image_set)
        image_sets_dict[polarisation] = _image_sets

    merged_image_set = task_merge_image_sets.submit(image_sets=image_sets_list)

    common_beam_shape = task_get_common_beam_from_image_set.submit(
        image_set=merged_image_set,
        cutoff=pol_field_options.beam_cutoff,
        fixed_beam_shape=pol_field_options.fixed_beam_shape,
    )

    stokes_beam_cubes: dict[str, list[PrefectFuture[Path]]] = {}
    for polarisation, image_set_list in image_sets_dict.items():
        with tags(f"polarisation-{polarisation}"):
            # Get the individual Stokes parameters in case of joint imaging
            if polarisation not in POLARISATION_MAPPING.keys():
                raise ValueError(f"Unknown polarisation {polarisation}")
            stokes_list = list(POLARISATION_MAPPING[polarisation])
            for stokes in stokes_list:
                with tags(f"stokes-{stokes}"):
                    beam_cubes: list[PrefectFuture[Path]] = []
                    for image_set in image_set_list:
                        stokes_image_list = task_split_and_get_image_set.submit(
                            image_set=image_set,
                            get=stokes,
                            by="pol",
                            mode="image",
                        )
                        convolved_image_list = task_convolve_images.submit(
                            image_paths=stokes_image_list,
                            beam_shape=common_beam_shape,
                            cutoff=pol_field_options.beam_cutoff,
                        )
                        # TODO: Consider accerating this by doing a linmos per-channel, then combining
                        channel_image_list = task_get_channel_images_from_paths.submit(
                            paths=convolved_image_list
                        )
                        prefix = task_getattr.submit(image_set, "prefix")

                        if polarisation == "linear":
                            # Get single Stokes prefix - the original prefix is the linear prefix
                            # i.e. `.qu.` -> `.q.` or `.u.` depending on the stokes
                            prefix = task_rename_linear_to_stokes.submit(
                                linear_name=prefix,
                                stokes=stokes,
                            )
                        cube_path = task_combine_images_to_cube.submit(
                            images=channel_image_list,
                            prefix=prefix,
                            mode="image",
                            remove_original_images=True,
                        )
                        beam_cubes.append(cube_path)
                stokes_beam_cubes[stokes] = beam_cubes

    linmos_result_list: list[PrefectFuture[LinmosResult]] = []
    # We run linmos now to ensure we have Stokes I images for leakage correction
    # If we have not imaged Stokes I, we cannot do leakage correction
    force_remove_leakage: bool | None = None
    if "i" not in stokes_beam_cubes.keys():
        force_remove_leakage = False

    linmos_options = task_create_object(
        object=LinmosOptions,
        holofile=pol_field_options.holofile,
        cutoff=pol_field_options.pb_cutoff,
        stokesi_images=stokes_beam_cubes.get("i"),
        force_remove_leakage=force_remove_leakage,
        trim_linmos_fits=pol_field_options.trim_linmos_fits,
    )
    for stokes, beam_cubes in stokes_beam_cubes.items():
        with tags(f"stokes-{stokes}"):
            linmos_result = task_linmos_images.submit(
                image_list=beam_cubes,
                container=pol_field_options.yandasoft_container,
                linmos_options=linmos_options,
                field_summary=field_summary,
            )
            linmos_result_list.append(linmos_result)

    # wait for all linmos results to be completed
    _ = [linmos_result.result() for linmos_result in linmos_result_list]


def setup_run_process_science_field(
    cluster_config: str | Path,
    flint_ms_directory: Path,
    pol_field_options: PolFieldOptions,
) -> None:
    science_sbid = get_sbid_from_path(path=flint_ms_directory)

    if pol_field_options.sbid_copy_path:
        updated_sbid_copy_path = pol_field_options.sbid_copy_path / f"{science_sbid}"
        logger.info(f"Updating archive copy path to {updated_sbid_copy_path=}")
        pol_field_options = pol_field_options.with_options(
            sbid_copy_path=updated_sbid_copy_path
        )

    dask_task_runner = get_dask_runner(cluster=cluster_config)

    process_science_fields_pol.with_options(
        name=f"Flint Polarisation Pipeline - {science_sbid}",
        task_runner=dask_task_runner,
    )(
        flint_ms_directory=flint_ms_directory,
        pol_field_options=pol_field_options,
    )


def get_parser() -> ArgumentParser:
    parser = ArgumentParser(description=__doc__)

    parser.add_argument(
        "--cli-config", is_config_file=True, help="Path to configuration file"
    )

    parser.add_argument(
        "flint_ms_directory",
        type=Path,
        help="Path to directories containing the beam-wise flint-calibrated MeasurementSets.",
    )
    parser.add_argument(
        "--cluster-config",
        type=str,
        default="petrichor",
        help="Path to a cluster configuration file, or a known cluster name. ",
    )

    parser = add_options_to_parser(
        parser=parser,
        options_class=PolFieldOptions,
        description="Polarisation processing options",
    )

    return parser


def cli() -> None:
    import logging

    # logger = logging.getLogger("flint")
    logger.setLevel(logging.INFO)

    parser = get_parser()

    args = parser.parse_args()

    field_options = create_options_from_parser(
        parser_namespace=args,
        options_class=PolFieldOptions,
    )

    setup_run_process_science_field(
        cluster_config=args.cluster_config,
        flint_ms_directory=args.flint_ms_directory,
        pol_field_options=field_options,
    )


if __name__ == "__main__":
    cli()



"""This is an interface into the yandasoft linmos task."""

from __future__ import annotations

from argparse import ArgumentParser
from pathlib import Path
from typing import Collection, Literal, NamedTuple

import numpy as np
from astropy.io import fits

from flint.logging import logger
from flint.naming import LinmosNames, create_linmos_names, extract_beam_from_name
from flint.options import BaseOptions, add_options_to_parser, create_options_from_parser
from flint.sclient import run_singularity_command
from flint.utils import remove_files_folders

# This is the expected orientation of the third-axis and footprint (remember the footprint
# can be electronically rotated).
EXPECTED_HOLOGRAPHY_ROTATION_CONSTANT_RADIANS = -2 * np.pi / 8


class LinmosResult(BaseOptions):
    cmd: str
    """The yandasoft linmos task that will be executed"""
    parset: Path
    """The output location that the generated linmos parset has been written to"""
    image_fits: Path
    """Path to the output linmos image created (or will be). """
    weight_fits: Path
    """Path to the output weight image formed through the linmos process"""


class LinmosOptions(BaseOptions):
    """Container for options that direct linmos processing"""

    base_output_name: Path = Path("linmos_field").absolute()
    "Base name path of the output linmos produces, including parent path. This is provided to form ``LinmosNames``. Defaults to 'linmos_field'."
    holofile: Path | None = None
    """Path to a FITS cube produced by the holography processing pipeline. Used by linmos to appropriate primary-beam correct the images. Defaults to None."""
    cutoff: float = 0.001
    """Pixels whose primary beam attenuation is below this cutoff value are blanked. Defaults to 0.001."""
    pol_axis: float | None = None
    """The physical oritentation of the ASKAP third-axis in radians. Defaults to None."""
    trim_linmos_fits: bool = True
    """Attempt to trim the output linmos files of as much empty space as possible. Defaults to True."""
    cleanup: bool = False
    """Remove files generated throughout linmos, including the text files with the channel weights. Defaults to False."""
    stokesi_images: list[Path] | None = None
    """Collection of Stokes-I images used to correct widefield leakage when input images are Stokes Q or U. If None ``linmos`` will attempt to find them. Defaults to None."""
    force_remove_leakage: bool | None = None
    """Overwrite the ``removeLeakage`` option to ``linmos``. Defaults to None. """
    overwrite: bool = False
    """Overwrite the linmos parset file generated if it exists. Defaults to False."""
    remove_original_images: bool = False
    """Delete the images that were coaddede together. Defaults to False."""


class BoundingBox(NamedTuple):
    """Simple container to represent a bounding box"""

    xmin: int
    """Minimum x pixel"""
    xmax: int
    """Maximum x pixel"""
    ymin: int
    """Minimum y pixel"""
    ymax: int
    """Maximum y pixel"""
    original_shape: tuple[int, int]
    """The original shape of the image. If constructed against a cube this is the shape of a single plane."""


class LinmosParsetSummary(NamedTuple):
    """Container for key components around a linmos parset file"""

    parset_path: Path
    """Path to the parset text file created"""
    image_paths: tuple[Path, ...]
    """The set of paths to the fits images that were coadded together"""
    weight_text_paths: tuple[Path, ...] | None = None
    """The set of Paths to the text files with per channel weights used by linmos"""


def _create_bound_box_plane(
    image_data: np.ndarray, is_masked: bool = False
) -> BoundingBox | None:
    """Create a bounding box around pixels in a 2D image. If all
    pixels are not valid, then ``None`` is returned.

    Args:
        image_data (np.ndarray): The 2D image to construct a bounding box around
        is_masked (bool, optional): Whether to treat the image as booleans or values. Defaults to False.

    Returns:
        Optional[BoundingBox]: None if no valid pixels, a bounding box with the (xmin,xmax,ymin,ymax) of valid pixels
    """
    assert len(image_data.shape) == 2, (
        f"Only two-dimensional arrays supported, received {image_data.shape}"
    )

    # First convert to a boolean array
    image_valid = image_data if is_masked else np.isfinite(image_data)

    if not any(image_valid.reshape(-1)):
        logger.info("No pixels to creating bounding box for")
        return None

    # Then make them 1D arrays
    x_valid = np.any(image_valid, axis=1)
    y_valid = np.any(image_valid, axis=0)

    # Now get the first and last index
    xmin, xmax = np.where(x_valid)[0][[0, -1]]
    ymin, ymax = np.where(y_valid)[0][[0, -1]]

    return BoundingBox(
        xmin=xmin, xmax=xmax, ymin=ymin, ymax=ymax, original_shape=image_data.shape[-2:]
    )


def create_bound_box(image_data: np.ndarray, is_masked: bool = False) -> BoundingBox:
    """Construct a bounding box around finite pixels for a 2D image.

    If a cube ids provided, the bounding box is constructed from pixels
    as broadcast across all of the non-spatial dimensions.  That is to
    say the single bounding box can be projected across all channel/stokes
    channels

    If ``is_mask` is ``False``, the ``image_data`` will be masked internally using ``numpy.isfinite``.

    Args:
        image_data (np.ndarray): The image data that will have a bounding box constructed for.
        is_masked (bool, optional): if this is ``True`` the ``image_data`` are treated as a boolean mask array. Defaults to False.

    Returns:
        BoundingBox: The tight bounding box around pixels.
    """
    reshaped_image_data = image_data.reshape((-1, *image_data.shape[-2:]))
    logger.info(f"New image shape {reshaped_image_data.shape} from {image_data.shape}")

    bounding_boxes = [
        _create_bound_box_plane(image_data=image, is_masked=is_masked)
        for image in reshaped_image_data
    ]
    bounding_boxes = [bb for bb in bounding_boxes if bb is not None]

    if len(bounding_boxes) == 0:
        logger.info("No valid bounding box found. Constructing one for all pixels")
        return BoundingBox(
            xmin=0,
            xmax=image_data.shape[-1] - 1,
            ymin=0,
            ymax=image_data.shape[-2] - 1,
            original_shape=tuple(image_data.shape[-2:]),  # type: ignore
        )
    elif len(bounding_boxes) == 1:
        assert bounding_boxes[0] is not None, "This should not happen"
        return bounding_boxes[0]

    assert all([bb is not None for bb in bounding_boxes])

    logger.info(
        f"Boounding boxes across {len(bounding_boxes)} constructed. Finsing limits. "
    )
    # The type ignores below are to avoid mypy believe bound_boxes could
    # include None. The above checks should be sufficient
    xmin = min([bb.xmin for bb in bounding_boxes])  # type: ignore
    xmax = max([bb.xmax for bb in bounding_boxes])  # type: ignore
    ymin = min([bb.ymin for bb in bounding_boxes])  # type: ignore
    ymax = max([bb.ymax for bb in bounding_boxes])  # type: ignore

    return BoundingBox(
        xmin=xmin, xmax=xmax, ymin=ymin, ymax=ymax, original_shape=image_data.shape
    )


class TrimImageResult(NamedTuple):
    """The constructed path and the bounding box"""

    path: Path
    """The path to the trimmed image"""
    bounding_box: BoundingBox
    """The bounding box that was applied to the image"""


def trim_fits_image(
    image_path: Path, bounding_box: BoundingBox | None = None
) -> TrimImageResult:
    """Trim the FITS image produces by linmos to remove as many empty pixels around
    the border of the image as possible. This is an inplace operation. Pixels
    that have a value of 0.0 are first filled with nan values before trimming.

    Args:
        image_path (Path): The FITS image that will have its border trimmed
        bounding_box (Optional[BoundingBox], optional): The bounding box that will be applied to the image. If None it is computed. Defaults to None.

    Returns:
        Path: Path of the FITS image that had its border trimmed
    """
    logger.info(f"Trimming {image_path.name}")
    with fits.open(image_path) as fits_image:
        data = fits_image[0].data  # type: ignore
        logger.info(f"Original data shape: {data.shape}")

        # 0.0 is not a real number, blank them out so the border
        # can be computed and trimmed correctly.
        logger.info("Blanking pixels with values of 0.0.")
        data[data == 0.0] = np.nan

        image_shape = data.shape[-2:]
        logger.info(f"The image dimensions are: {image_shape}")

        if not bounding_box:
            logger.info("Constructing a new bounding box")
            bounding_box = create_bound_box(
                image_data=np.squeeze(data), is_masked=False
            )
            logger.info(f"Constructed {bounding_box=}")
        else:
            if image_shape != bounding_box.original_shape[-2:]:
                raise ValueError(
                    f"Bounding box constructed against {bounding_box.original_shape}, but being applied to {image_shape=}"
                )

        data = data[
            ...,
            bounding_box.xmin : bounding_box.xmax,
            bounding_box.ymin : bounding_box.ymax,
        ]

        header = fits_image[0].header  # type: ignore
        header["CRPIX1"] -= bounding_box.ymin
        header["CRPIX2"] -= bounding_box.xmin

        logger.info(f"Trimmed data shape: {data.shape}")

    fits.writeto(filename=image_path, data=data, header=header, overwrite=True)

    return TrimImageResult(path=image_path, bounding_box=bounding_box)


def _get_image_weight_plane(
    image_data: np.ndarray, mode: Literal["std", "mad"] = "mad", stride: int = 4
) -> float:
    """Extract the inverse variance weight for an input plane of data

    Modes are 'std' or 'mad'.

    Args:
        image_data (np.ndarray): Data to consider
        mode (str, optional): Statistic computation mode. Defaults to "mad".
        stride (int, optional): Include every n'th pixel when computing the weight. '1' includes all pixels. Defaults to 1.

    Raises:
        ValueError: Raised when modes unknown

    Returns:
        float: The inverse variance weight computerd
    """

    weight_modes = ("mad", "std")
    assert mode in weight_modes, (
        f"Invalid {mode=} specified. Available modes: {weight_modes}"
    )

    # remove non-finite numbers that would ruin the statistic
    image_data = image_data[np.isfinite(image_data)][::stride]

    if np.all(~np.isfinite(image_data)):
        return 0.0

    if mode == "mad":
        median = np.median(image_data)
        mad = np.median(np.abs(image_data - median))
        weight = 1.0 / mad**2
    elif mode == "std":
        std = np.std(image_data)
        weight = 1.0 / std**2
    else:
        raise ValueError(f"Invalid {mode=} specified. Available modes: {weight_modes}")

    float_weight = float(weight)
    return float_weight if np.isfinite(float_weight) else 0.0


def get_image_weight(
    image_path: Path, mode: str = "mad", stride: int = 1, image_slice: int = 0
) -> list[float]:
    """Compute an image weight supplied to linmos, which is used for optimally
    weighting overlapping images. Supported modes are 'mad' and 'mtd', which
    simply resolve to their numpy equivalents.

    This weight is really a relative weight to used between all images in a set
    of images being co-added together. So long as these are all calculated in
    the same way, it does not necessarily have to correspond to an optimatelly
    calculated RMS.

    The stride parameter will only include every N'th pixel when computing the
    weights. A smaller set of pixels will reduce the time required to calculate
    the weights, but may come at the cost of accuracy with large values.

    Args:
        image (Path): The path to the image fits file to inspect.
        mode (str, optional): Which mode should be used when calculating the weight. Defaults to 'mad'.
        stride (int, optional): Include every n'th pixel when computing the weight. '1' includes all pixels. Defaults to 1.
        image_slice (int, optional): The image slice in the HDU list of the `image` fits file to inspect. Defaults to 0.

    Raises:
        ValueError: Raised when a mode is requested but does not exist

    Returns:
        List[float]: The weight per channel to supply to linmos
    """

    logger.debug(
        f"Computing linmos weight using {mode=}, {image_slice=} for {image_path}. "
    )

    weights: list[float] = []
    with fits.open(image_path, memmap=True) as in_fits:
        image_data = in_fits[image_slice].data  # type: ignore

        assert len(image_data.shape) >= 2, (
            f"{len(image_data.shape)=} is less than two. Is this really an image?"
        )

        image_shape = image_data.shape[-2:]
        image_data = (
            image_data.reshape((-1, *image_shape))
            if len(image_data.shape)
            else image_data
        )

        assert len(image_data.shape) == 3, (
            f"Expected to have shape (chan, dec, ra), got {image_data.shape}"
        )

        for idx, chan_image_data in enumerate(image_data):
            weight = _get_image_weight_plane(image_data=chan_image_data, stride=stride)
            logger.info(f"Channel {idx} {weight=:.3f} for {image_path}")

            weights.append(weight)

    return weights


def generate_weights_list_and_files(
    image_paths: Collection[Path], mode: str = "mad", stride: int = 1
) -> tuple[Path, ...]:
    """Generate the expected linmos weight files, and construct an appropriate
    string that can be embedded into a linmos partset. These weights files will
    appear as:

    >>> #Channel Weight
    >>> 0 1234.5
    >>> 1 6789.0

    The weights should be correct relative to the entire set of input images.
    They do not necessarily have to correspond to an accurate measure of the RMS.

    This function will create a corresponding text file for each input image. At
    the moment it is only intended to work on MFS images. It *is not* currently
    intended to be used on image cubes.

    The stride parameter will only include every N'th pixel when computing the
    weights. A smaller set of pixels will reduce the time required to calculate
    the weights, but may come at the cost of accuracy with large values.

    Args:
        image_paths (Collection[Path]): Images to iterate over to create a corresponding weights.txt file.
        mode (str, optional): The mode to use when calling get_image_weight

    Returns:
        Tuple[Path, ...]: A list of paths pointing to the weights for each input image
    """
    logger.info(
        f"No weights provided. Calculating weights for {len(image_paths)} images."
    )

    # TODO: image cubes should be supported here. This would required iterating
    # over each channel in the FITS cube.
    weight_file_list = []
    for image in image_paths:
        weight_file = image.with_suffix(".weights.txt")
        weight_file_list.append(weight_file)

        # Must be of the format:
        # #Channel Weight
        # 0 1234.5
        # 1 6789.0
        with open(weight_file, "w") as out_file:
            logger.info(f"Writing {weight_file}")
            out_file.write("#Channel Weight\n")
            image_weights = get_image_weight(image_path=image, mode=mode, stride=stride)
            weights = "\n".join(
                [f"{idx} {weight}" for idx, weight in enumerate(image_weights)]
            )
            out_file.write(weights)
            out_file.write("\n")  # Required for linmos to properly process weights

    return tuple(weight_file_list)


def _get_alpha_linmos_option(pol_axis: float | None = None) -> str:
    """Compute the appropriate alpha term for linmos that is used to
    describe the differential rotation of the ASKAP third-axis and the
    footprint layout. The typical holography rotation is -45 degs. Internally
    the `alpha` term is computed as:

    >>> pol_axis - EXPECTED_HOLOGRAPHY_ROTATION_CONSTANT_RADIANS

    Args:
        pol_axis (Optional[float], optional): The prescribed polarisation axis value described in a MS. Defaults to None.

    Returns:
        str: Yandasoft linmos option to rotation the holography cubes
    """

    # Work out the appropriate rotation to provide linmos. This should
    # be the differential pol. axis. rotation between the science field
    # and the holography. At the moment this holography rotation is
    # unknown to us (not stored in the cube header).
    if pol_axis is None:
        return ""

    assert np.abs(pol_axis) <= 2.0 * np.pi, (
        f"{pol_axis=}, which is outside +/- 2pi radians and seems unreasonable"
    )

    logger.info(
        f"The constant assumed holography rotation is: {EXPECTED_HOLOGRAPHY_ROTATION_CONSTANT_RADIANS:.4f} radians"
    )
    logger.info(f"The extracted pol_axis of the field: {pol_axis:.4f} radians")
    alpha = pol_axis - EXPECTED_HOLOGRAPHY_ROTATION_CONSTANT_RADIANS
    logger.info(f"Differential rotation is: {alpha} rad")

    return f"linmos.primarybeam.ASKAP_PB.alpha = {alpha} # in radians\n"


def _get_holography_linmos_options(
    holofile: Path | None = None,
    pol_axis: float | None = None,
    remove_leakage: bool = False,
    stokesi_images: Collection[Path] | None = None,
) -> str:
    """Construct the appropriate set of linmos options that
    describe the use of the holography cube file to primary
    beam correct the input images. This includes appropriately
    rotating the holography (see `_get_alpha_linmos_options`).

    Args:
        holofile (Optional[Path], optional): Path to the holography cube file to primary beam correct with. Defaults to None.
        pol_axis (Optional[float], optional): The rotation of the third axis as described in an ASAKP MS. Defaults to None.
        remove_leakage (bool, optional): Add the directive to remove leakage. Defaults to False.

    Returns:
        str: Set of linmos options to add to a parset file
    """

    if holofile is None:
        return ""

    # This requires an appropriate holography fits cube to
    # have been created. This is typically done outside as
    # part of operations.
    logger.info(f"Using holography file {holofile} -- setting removeleakge to true")
    assert holofile.exists(), f"{holofile=} has been specified but does not exist. "

    parset = (
        f"linmos.primarybeam      = ASKAP_PB\n"
        f"linmos.primarybeam.ASKAP_PB.image = {holofile.absolute()!s}\n"
        f"linmos.removeleakage    = {'true' if remove_leakage else 'false'}\n"
    )
    parset += _get_alpha_linmos_option(pol_axis=pol_axis)

    if stokesi_images is not None:
        logger.info("Stokes I images provided. Adding to linmos parset.")
        stokesi_list = _file_list_to_string(stokesi_images)
        parset += f"linmos.stokesinames = {stokesi_list}\n"

    return parset


def _file_list_to_string(file_list: Collection[Path]) -> str:
    img_str: list[str] = list(
        [str(p).replace(".fits", "") for p in file_list if p.exists()]
    )
    logger.info(
        f"{len(img_str)} unique images from {len(file_list)} input collection. "
    )
    img_list: str = "[" + ",".join(img_str) + "]"

    assert len(set(img_str)) == len(file_list), (
        f"Some images were dropped from the linmos image string (found {len(set(img_str))}, expcected {len(file_list)}). Walk the plank. "
    )

    return img_list


def generate_linmos_parameter_set(
    images: Collection[Path],
    linmos_names: LinmosNames,
    linmos_options: LinmosOptions,
    weight_list: str | None = None,
) -> LinmosParsetSummary:
    """Generate a parset file that will be used with the
    yandasoft linmos task.

    Args:
        images (Collection[Path]): The images that will be coadded into a single field image.
        linmos_names (LinmosNames): Names of the output image and weights that linmos will produces. The weight image will have a similar name. Defaults to "linmos_field".
        linmos_options (LinmosOptions): Options that are passed through to the yandasoft linmos application.
        weight_list (str, optional): If not None, this string will be embedded into the yandasoft linmos parset as-is. It should represent the formatted string pointing to weight files, and should be equal length of the input images. If None it is internally generated. Defaults to None.

    Returns:
        LinmosParsetSummary: Important components around the generated parset file.
    """

    img_list = _file_list_to_string(images)

    if linmos_options.stokesi_images is not None and len(
        linmos_options.stokesi_images
    ) != len(images):
        raise ValueError(
            f"Stokes I images provided {len(linmos_options.stokesi_images)} do not match the number of input images {len(images)}"
        )

    # If no weights_list has been provided (and therefore no optimal
    # beam-wise weighting) assume that all beams are of about the same
    # quality. In reality, this should be updated to provide a RMS noise
    # estimate per-pixel of each image.
    weight_str = weight_list
    weight_files: tuple[Path, ...] | None = None
    if weight_str is None:
        weight_files = generate_weights_list_and_files(
            image_paths=images, mode="mad", stride=8
        )
        assert weight_files is not None, (
            f"{weight_files=}, which should not happen after creating weight files"
        )
        weight_str = _file_list_to_string(weight_files)

    beam_order_strs = [str(extract_beam_from_name(str(p.name))) for p in images]
    beam_order_list = "[" + ",".join(beam_order_strs) + "]"

    # The yandasoft linmos tasks will insist on adding a .fits extension
    # so it needs to be dropped from the Path objects. Using the Path.stem
    # attribute drops the parent directory (absolute directory).
    parent_dir = linmos_names.image_fits.parent

    # TODO: This should be a list of one line strings that is grown based on
    # options provided, then joined with a "\n".join(parset)
    # Parameters are taken from arrakis
    parset = (
        f"linmos.names            = {img_list}\n"
        f"linmos.weights          = {weight_str}\n"
        f"linmos.beams            = {beam_order_list}\n"
        # f"linmos.beamangle        = {beam_angle_list}\n"
        f"linmos.imagetype        = fits\n"
        f"linmos.outname          = {parent_dir / linmos_names.image_fits.stem!s}\n"
        f"linmos.outweight        = {parent_dir / linmos_names.weight_fits.stem!s}\n"
        f"# For ASKAPsoft>1.3.0\n"
        f"linmos.useweightslog    = true\n"
        f"linmos.weighttype       = Combined\n"
        f"linmos.weightstate      = Inherent\n"
        f"linmos.cutoff           = 0\n"  # This `cutoff` is based on weights, not primary beam attenuation
        f"linmos.finalcutoff           = {linmos_options.cutoff}\n"  # This one though, uses the PB.
    )
    # Construct the holography section of the linmos parset
    remove_leakage = (linmos_options.holofile is not None) and (
        ".i." not in str(next(iter(images)))
    )
    if linmos_options.force_remove_leakage is not None:
        logger.info(
            f"Force removing leakage: Setting to {linmos_options.force_remove_leakage}"
        )
        remove_leakage = linmos_options.force_remove_leakage

    parset += _get_holography_linmos_options(
        holofile=linmos_options.holofile,
        pol_axis=linmos_options.pol_axis,
        remove_leakage=remove_leakage,
        stokesi_images=linmos_options.stokesi_images,
    )

    # Now write the file, me hearty
    logger.info(f"Writing parset to {linmos_names.parset_output_path!s}.")
    logger.info(f"{parset}")
    if not linmos_options.overwrite:
        assert not Path(linmos_names.parset_output_path).exists(), (
            f"The parset {linmos_names.parset_output_path} already exists!"
        )
    with open(linmos_names.parset_output_path, "w") as parset_file:
        parset_file.write(parset)

    linmos_parset_summary = LinmosParsetSummary(
        parset_path=linmos_names.parset_output_path,
        weight_text_paths=tuple(map(Path, weight_files))
        if weight_files
        else weight_files,
        image_paths=tuple(map(Path, images)),
    )

    return linmos_parset_summary


def _linmos_cleanup(linmos_parset_summary: LinmosParsetSummary) -> tuple[Path, ...]:
    """Clean up linmos files if requested.

    Args:
        linmos_parset_summary (LinmosParsetSummary): Parset summary from which the text file weights are gathered for deletion from

    Returns:
        Tuple[Path, ...]: Set of files removed
    """

    from flint.utils import remove_files_folders

    removed_files = []
    if linmos_parset_summary.weight_text_paths is not None:
        removed_files.extend(
            remove_files_folders(*linmos_parset_summary.weight_text_paths)
        )
    return tuple(removed_files)


# TODO: These options are starting to get a little large. Perhaps we should use BaseOptions.
def linmos_images(
    images: Collection[Path],
    linmos_options: LinmosOptions,
    parset_output_path: Path | None = None,
    weight_list: str | None = None,
    container: Path = Path("yandasoft.sif"),
) -> LinmosResult:
    """Create a linmos parset file and execute it.

    Args:
        images (Collection[Path]): The images that will be coadded into a single field image.
        linmos_options (LinmosOptions): Options to control the yandasott linmos program and related features.
        parset_output_path (Path | None, optional): Path of the output linmos parset file. If None it is derived from common input fields. Defaults to None.
        weight_list (str, optional): If not None, this string will be embedded into the yandasoft linmos parset as-is. It should represent the formatted string pointing to weight files, and should be equal length of the input images. If None it is internally generated. Defaults to None.

    Returns:
        LinmosResult: The linmos command executed and the associated parset file
    """

    assert container.exists(), f"The yandasoft container {container!s} was not found. "

    linmos_names: LinmosNames = create_linmos_names(
        name_prefix=linmos_options.base_output_name,
        parset_output_path=parset_output_path,
    )

    linmos_parset_summary = generate_linmos_parameter_set(
        images=images,
        linmos_names=linmos_names,
        linmos_options=linmos_options,
        weight_list=weight_list,
    )

    linmos_cmd_str = f"linmos -c {linmos_parset_summary.parset_path!s}"
    bind_dirs = [image.absolute().parent for image in images] + [
        linmos_parset_summary.parset_path.absolute().parent
    ]
    if linmos_options.holofile:
        bind_dirs.append(linmos_options.holofile.absolute().parent)

    run_singularity_command(
        image=container, command=linmos_cmd_str, bind_dirs=bind_dirs
    )

    linmos_result = LinmosResult(
        cmd=linmos_cmd_str,
        parset=linmos_parset_summary.parset_path,
        image_fits=linmos_names.image_fits.absolute(),
        weight_fits=linmos_names.weight_fits.absolute(),
    )

    # Trim the fits image to remove empty pixels
    if linmos_options.trim_linmos_fits:
        image_trim_results = trim_fits_image(image_path=linmos_names.image_fits)
        trim_fits_image(
            image_path=linmos_names.weight_fits,
            bounding_box=image_trim_results.bounding_box,
        )

    if linmos_options.cleanup:
        _linmos_cleanup(linmos_parset_summary=linmos_parset_summary)

    if linmos_options.remove_original_images:
        remove_files_folders(*images)

    return linmos_result


def get_parser() -> ArgumentParser:
    parser = ArgumentParser(description=__doc__)

    subparsers = parser.add_subparsers(dest="mode")

    parset_parser = subparsers.add_parser(
        "parset", help="Generate a yandasoft linmos parset"
    )

    parset_parser.add_argument(
        "images", type=Path, nargs="+", help="The images that will be coadded"
    )
    parset_parser.add_argument(
        "parset_output_path", type=Path, help="The output path of the linmos parset"
    )
    parset_parser.add_argument(
        "--weight-list",
        type=Path,
        default=None,
        help="Path a new-line delimited text-file containing the relative weights corresponding to the input images",
    )
    parset_parser.add_argument(
        "--yandasoft-container",
        type=Path,
        default=None,
        help="Path to the container with yandasoft tools",
    )
    parset_parser = add_options_to_parser(
        parser=parset_parser, options_class=LinmosOptions
    )

    trim_parser = subparsers.add_parser(
        "trim", help="Remove blank border of FITS image"
    )

    trim_parser.add_argument(
        "images", type=Path, nargs="+", help="The images that will be trimmed"
    )
    return parser


def cli() -> None:
    parser = get_parser()

    args = parser.parse_args()

    if args.mode == "parset":
        linmos_options = create_options_from_parser(
            parser_namespace=args, options_class=LinmosOptions
        )
        if args.yandasoft_container is None:
            linmos_names = create_linmos_names(
                name_prefix=args.base_output_name,
                parset_output_path=args.parset_output_path,
            )
            generate_linmos_parameter_set(
                images=args.images,
                linmos_names=linmos_names,
                linmos_options=linmos_options,
                weight_list=args.weight_list,
            )
        else:
            linmos_images(
                images=args.images,
                parset_output_path=args.parset_output_path,
                weight_list=args.weight_list,
                linmos_options=linmos_options,
                container=args.yandasoft_container,
            )
    elif args.mode == "trim":
        images = args.images
        logger.info(f"Will be trimming {len(images)}")
        for image in images:
            trim_fits_image(image_path=Path(image))
    else:
        logger.error(f"Unrecognised mode: {args.mode}")


if __name__ == "__main__":
    cli()



"""Test utilities related to flagging measurement set operations"""

from __future__ import annotations

import shutil
from pathlib import Path

import numpy as np
import pytest
from casacore.tables import table

from flint.flagging import flag_ms_zero_uvws, nan_zero_extreme_flag_ms
from flint.utils import get_packaged_resource_path


@pytest.fixture
def ms_example(tmpdir):
    ms_zip = Path(
        get_packaged_resource_path(
            package="flint.data.tests",
            filename="SB39400.RACS_0635-31.beam0.small.ms.zip",
        )
    )
    outpath = Path(tmpdir) / "39400"

    shutil.unpack_archive(ms_zip, outpath)

    ms_path = Path(outpath) / "SB39400.RACS_0635-31.beam0.small.ms"

    return ms_path


def test_flag_ms_zero_uvws(ms_example):
    flag_ms_zero_uvws(ms=ms_example)

    with table(str(ms_example), ack=False) as tab:
        uvws = tab.getcol("UVW")
        flags = tab.getcol("FLAG")

        uvw_mask = np.all(uvws == 0, axis=1)

        assert np.all(flags[uvw_mask] == True)  # noQA: E712


def test_add_flag_ms_zero_uvws(ms_example):
    with table(str(ms_example), readonly=False, ack=False) as tab:
        uvws = tab.getcol("UVW")
        uvws[:1, :] = 0

        tab.putcol("UVW", uvws)

    flag_ms_zero_uvws(ms=ms_example)

    with table(str(ms_example), ack=False) as tab:
        uvws = tab.getcol("UVW")

        flags = tab.getcol("FLAG")

        uvw_mask = np.all(uvws == 0, axis=1)

        assert np.sum(uvw_mask) > 0
        assert np.all(flags[uvw_mask] == True)  # noQA: E712


def test_nan_zero_extreme_flag_ms(ms_example):
    """Makes sure that flagging the NaNs, UVWs of zero and
    extreme outliers works. Was added after introducing the
    chunking"""
    with table(str(ms_example), readonly=False, ack=False) as tab:
        uvws = tab.getcol("UVW")
        uvws[:10, :] = 0

        tab.putcol("UVW", uvws)

    nan_zero_extreme_flag_ms(ms=ms_example)

    with table(str(ms_example), ack=False) as tab:
        uvws = tab.getcol("UVW")

        flags = tab.getcol("FLAG")

        uvw_mask = np.all(uvws == 0, axis=1)

        assert np.sum(uvw_mask) > 0
        assert np.all(flags[uvw_mask] == True)  # noQA: E712


def test_nan_zero_extreme_flag_ms_with_chunks(ms_example):
    """Makes sure that flagging the NaNs, UVWs of zero and
    extreme outliers works. Was added after introducing the
    chunking.

    Same as above test but with chunk size"""
    with table(str(ms_example), readonly=False, ack=False) as tab:
        uvws = tab.getcol("UVW")
        uvws[:20, :] = 0

        tab.putcol("UVW", uvws)

    nan_zero_extreme_flag_ms(ms=ms_example, chunk_size=1)

    with table(str(ms_example), ack=False) as tab:
        uvws = tab.getcol("UVW")

        flags = tab.getcol("FLAG")

        uvw_mask = np.all(uvws == 0, axis=1)

        assert np.sum(uvw_mask) > 0
        assert np.all(flags[uvw_mask] == True)  # noQA: E712


def test_nan_zero_extreme_flag_ms_with_chunks_and_datanan(ms_example):
    """Makes sure that flagging the NaNs, UVWs of zero and
    extreme outliers works. Was added after introducing the
    chunking.

    Same as above test but with chunk size and naning data when flags are true"""
    with table(str(ms_example), readonly=False, ack=False) as tab:
        uvws = tab.getcol("UVW")
        uvws[:20, :] = 0
        original_data = tab.getcol("DATA")
        tab.putcol("UVW", uvws)

    nan_zero_extreme_flag_ms(ms=ms_example, chunk_size=1, nan_data_on_flag=True)

    with table(str(ms_example), ack=False) as tab:
        uvws = tab.getcol("UVW")

        flags = tab.getcol("FLAG")
        data = tab.getcol("DATA")
        uvw_mask = np.all(uvws == 0, axis=1)

        assert np.sum(np.isnan(data)) == np.sum(flags)
        assert np.sum(np.isnan(data)) > np.sum(np.isnan(original_data))
        assert np.sum(uvw_mask) > 0
        assert np.all(flags[uvw_mask] == True)  # noQA: E712



"""Tests related to components in the yandasoft linmos coadd.
At the moment this is not testing the actual application. Just
some of the helper functions around it.
"""

from __future__ import annotations

from pathlib import Path

import numpy as np
import pytest
from astropy.io import fits

from flint.coadd.linmos import (
    BoundingBox,
    LinmosOptions,
    LinmosParsetSummary,
    _create_bound_box_plane,
    _get_alpha_linmos_option,
    _get_holography_linmos_options,
    _get_image_weight_plane,
    _linmos_cleanup,
    create_bound_box,
    generate_weights_list_and_files,
    trim_fits_image,
)
from flint.naming import create_linmos_base_path


def get_lots_of_names_2() -> list[Path]:
    examples = [
        "59058/SB59058.RACS_1626-84.round4.i.ch0285-0286.linmos.fits",
        "59058/SB59058.RACS_1626-84.round4.i.ch0285-0286.linmos.fits",
        "59058/SB59058.RACS_1626-84.round4.i.ch0070-0071.linmos.fits",
        "59058/SB59058.RACS_1626-84.round4.i.ch0142-0143.linmos.fits",
        "59058/SB59058.RACS_1626-84.round4.i.ch0214-0215.linmos.fits",
        "59058/SB59058.RACS_1626-84.round4.i.ch0286-0287.linmos.fits",
        "59058/SB59058.RACS_1626-84.round4.i.ch0071-0072.linmos.fits",
        "59058/SB59058.RACS_1626-84.round4.i.ch0143-0144.linmos.fits",
        "59058/SB59058.RACS_1626-84.round4.i.ch0215-0216.linmos.fits",
        "59058/SB59058.RACS_1626-84.round4.i.ch0287-0288.linmos.fits",
    ]

    return list(map(Path, examples))


def test_create_name_to_linmos_options():
    """Make sure that the base name created can be passed to an
    instance of LinmosOptions."""
    # Seems silly but was burnt before by not testing this incorrect type
    examples = get_lots_of_names_2()
    # These are sanity, pirates trust nothing
    common_names = create_linmos_base_path(input_images=examples)
    expected_common_name = Path("59058/SB59058.RACS_1626-84.round4.i").absolute()

    assert common_names == expected_common_name
    _ = LinmosOptions(base_output_name=common_names)


def test_get_image_weight_plane():
    """The extraction of weights per plane"""
    data = np.arange(100).reshape((10, 10))

    with pytest.raises(AssertionError):
        _get_image_weight_plane(image_data=data, mode="noexists")  # type: ignore

    assert np.isclose(
        0.0016,
        _get_image_weight_plane(image_data=data, mode="mad", stride=1),
        atol=0.0001,
    )
    assert np.isclose(
        0.00120012,
        _get_image_weight_plane(image_data=data, mode="std", stride=1),
        atol=0.0001,
    )

    data = np.arange(100).reshape((10, 10)) * np.nan
    assert _get_image_weight_plane(image_data=data) == 0.0


def create_fits_image(out_path, image_size=(1000, 1000), set_to_nan: bool = True):
    data = np.zeros(image_size)
    data[10:600, 20:500] = 1
    if set_to_nan:
        data[data == 0] = np.nan

    header = fits.header.Header({"CRPIX1": 10, "CRPIX2": 20})

    fits.writeto(out_path, data=data, header=header)


def create_image_cube(out_path):
    data = np.arange(20 * 100).reshape((20, 10, 10))
    header = fits.header.Header({"CRPIX1": 10, "CRPIX2": 20, "CRPIX3": 1})

    fits.writeto(out_path, header=header, data=data)


def test_linmos_alpha_option():
    """Ensure the rotation string supplied to linmos is calculated appropriately"""

    options_str = _get_alpha_linmos_option(pol_axis=None)
    assert options_str == ""

    options_str = _get_alpha_linmos_option(pol_axis=np.deg2rad(-45))
    expected_str = "linmos.primarybeam.ASKAP_PB.alpha = 0.0 # in radians\n"
    assert options_str == expected_str

    with pytest.raises(AssertionError):
        _get_alpha_linmos_option(pol_axis=1234)


def test_cleanup_image_weights_(tmpdir):
    """Remove the weight files that have been created"""
    cube_weight = Path(tmpdir) / "cubeweight"
    cube_weight.mkdir(parents=True, exist_ok=True)
    cube_fits = cube_weight / "cube.fits"

    create_image_cube(out_path=cube_fits)
    weight_file = cube_fits.with_suffix(".weights.txt")
    assert not weight_file.exists()

    weight_paths = generate_weights_list_and_files(image_paths=[cube_fits], mode="mad")
    linmos_parset_summary = LinmosParsetSummary(
        parset_path=Path("JackSparrow.txt"),
        image_paths=tuple([cube_fits]),
        weight_text_paths=weight_paths,
    )
    assert weight_file.exists()
    assert isinstance(linmos_parset_summary, LinmosParsetSummary)
    files_removed = _linmos_cleanup(linmos_parset_summary=linmos_parset_summary)
    assert not weight_file.exists()
    assert len(files_removed) == 1
    assert files_removed[0] == weight_file


def test_get_image_weights(tmpdir):
    """See whether the weights computed per plane in a cube work appropriately"""
    cube_weight = Path(tmpdir) / "cubeweight"
    cube_weight.mkdir(parents=True, exist_ok=True)
    cube_fits = cube_weight / "cube.fits"

    create_image_cube(out_path=cube_fits)
    weight_file = cube_fits.with_suffix(".weights.txt")
    assert not weight_file.exists()

    generate_weights_list_and_files(image_paths=[cube_fits], mode="mad")
    assert weight_file.exists()
    # The file must end with a newline for linmos to work
    lines = weight_file.read_text().split("\n")
    assert len(lines) == 22, f"{lines}"


def test_get_image_weight_with_strides(tmpdir):
    """See whether the weights computed per plane in a cube work appropriately when striding over data"""
    cube_weight = Path(tmpdir) / "cubeweight"
    cube_weight.mkdir(parents=True, exist_ok=True)
    cube_fits = cube_weight / "cube.fits"

    create_image_cube(out_path=cube_fits)
    weight_file = cube_fits.with_suffix(".weights.txt")
    assert not weight_file.exists()

    generate_weights_list_and_files(image_paths=[cube_fits], mode="mad", stride=10)
    assert weight_file.exists()
    # The file must end with a newline for linmos to work
    lines = weight_file.read_text().split("\n")
    assert len(lines) == 22, f"{lines}"


def test_linmos_holo_options(tmpdir):
    holofile = Path(tmpdir) / "testholooptions/holo_file.fits"
    holofile.parent.mkdir(parents=True, exist_ok=True)

    ifile = Path(tmpdir) / "blackpearl.fits"
    ifile.touch()
    ifile.parent.mkdir(parents=True, exist_ok=True)

    with pytest.raises(AssertionError):
        _get_holography_linmos_options(holofile=holofile, pol_axis=None)

    assert _get_holography_linmos_options(holofile=None, pol_axis=None) == ""

    with holofile.open("w") as f:
        f.write("test")

    parset = _get_holography_linmos_options(holofile=holofile, pol_axis=None)
    assert "linmos.primarybeam      = ASKAP_PB\n" in parset
    assert "linmos.removeleakage    = false\n" in parset
    assert f"linmos.primarybeam.ASKAP_PB.image = {holofile.absolute()!s}\n" in parset
    assert "linmos.primarybeam.ASKAP_PB.alpha" not in parset

    parset = _get_holography_linmos_options(holofile=holofile, pol_axis=np.deg2rad(-45))
    assert "linmos.primarybeam      = ASKAP_PB\n" in parset
    assert "linmos.removeleakage    = false\n" in parset
    assert f"linmos.primarybeam.ASKAP_PB.image = {holofile.absolute()!s}\n" in parset
    assert "linmos.primarybeam.ASKAP_PB.alpha" in parset

    parset = _get_holography_linmos_options(
        holofile=holofile, remove_leakage=True, pol_axis=np.deg2rad(-45)
    )
    assert "linmos.primarybeam      = ASKAP_PB\n" in parset
    assert "linmos.removeleakage    = true\n" in parset
    assert f"linmos.primarybeam.ASKAP_PB.image = {holofile.absolute()!s}\n" in parset
    assert "linmos.primarybeam.ASKAP_PB.alpha" in parset

    parset = _get_holography_linmos_options(
        holofile=holofile,
        remove_leakage=True,
        pol_axis=np.deg2rad(-45),
        stokesi_images=[
            ifile,
        ],
    )
    from flint.logging import logger

    logger.info(parset)
    assert "linmos.primarybeam      = ASKAP_PB\n" in parset
    assert "linmos.removeleakage    = true\n" in parset
    assert f"linmos.primarybeam.ASKAP_PB.image = {holofile.absolute()!s}\n" in parset
    assert "linmos.primarybeam.ASKAP_PB.alpha" in parset
    assert f"linmos.stokesinames = [{ifile.with_suffix('').as_posix()}]\n" in parset

    with pytest.raises(AssertionError):
        parset = _get_holography_linmos_options(
            holofile=holofile,
            remove_leakage=True,
            pol_axis=np.deg2rad(-45),
            stokesi_images=[
                Path("doesnotexist.fits"),
            ],
        )


def test_trim_fits_while_blanking(tmp_path):
    """Ensure that fits files can be trimmed appropriately based on row/columns with valid pixels.
    This will require pixels with values of 0.0 to be naned"""
    tmp_dir = tmp_path / "imagenan"
    tmp_dir.mkdir()

    out_fits = tmp_dir / "example.fits"

    create_fits_image(out_fits, set_to_nan=False)
    og_hdr = fits.getheader(out_fits)
    assert og_hdr["CRPIX1"] == 10
    assert og_hdr["CRPIX2"] == 20

    trim_fits_image(out_fits)
    trim_hdr = fits.getheader(out_fits)
    trim_data = fits.getdata(out_fits)
    assert trim_hdr["CRPIX1"] == -10
    assert trim_hdr["CRPIX2"] == 10
    assert trim_data.shape == (589, 479)
    assert np.sum(trim_data == 0.0) == 0


def test_trim_fits(tmp_path):
    """Ensure that fits files can be trimmed appropriately based on row/columns with valid pixels"""
    tmp_dir = tmp_path / "image"
    tmp_dir.mkdir()

    out_fits = tmp_dir / "example.fits"

    create_fits_image(out_fits)
    og_hdr = fits.getheader(out_fits)
    assert og_hdr["CRPIX1"] == 10
    assert og_hdr["CRPIX2"] == 20

    trim_fits_image(out_fits)
    trim_hdr = fits.getheader(out_fits)
    trim_data = fits.getdata(out_fits)
    assert trim_hdr["CRPIX1"] == -10
    assert trim_hdr["CRPIX2"] == 10
    assert trim_data.shape == (589, 479)


def test_trim_fits_cube(tmp_path):
    """Ensure that fits files that has cube can be trimmed appropriately based on row/columns with valid pixels"""
    tmp_dir = tmp_path / "cube"
    tmp_dir.mkdir()

    out_fits = tmp_dir / "example.fits"

    cube_size = (12, 1000, 1000)
    data = np.zeros(cube_size)
    data[3, 10:600, 20:500] = 1
    data[data == 0] = np.nan

    header = fits.header.Header({"CRPIX1": 10, "CRPIX2": 20})

    fits.writeto(out_fits, data=data, header=header)

    og_hdr = fits.getheader(out_fits)
    assert og_hdr["CRPIX1"] == 10
    assert og_hdr["CRPIX2"] == 20

    trim_fits_image(out_fits)
    trim_hdr = fits.getheader(out_fits)
    trim_data = fits.getdata(out_fits)
    assert trim_hdr["CRPIX1"] == -10
    assert trim_hdr["CRPIX2"] == 10
    assert trim_data.shape == (12, 589, 479)  # type: ignore


def test_trim_fits_image_matching(tmp_path):
    """See the the bounding box can be passed through for matching to cutout"""

    tmp_dir = Path(tmp_path) / "image_bb_match"
    tmp_dir.mkdir()

    out_fits = tmp_dir / "example.fits"

    create_fits_image(out_fits)
    og_trim = trim_fits_image(out_fits)

    out_fits2 = tmp_dir / "example2.fits"
    create_fits_image(out_fits2)
    og_hdr = fits.getheader(out_fits2)
    assert og_hdr["CRPIX1"] == 10
    assert og_hdr["CRPIX2"] == 20

    trim_fits_image(image_path=out_fits2, bounding_box=og_trim.bounding_box)
    trim_hdr = fits.getheader(out_fits2)
    trim_data = fits.getdata(out_fits2)
    assert trim_hdr["CRPIX1"] == -10
    assert trim_hdr["CRPIX2"] == 10
    assert trim_data.shape == (589, 479)

    out_fits2 = tmp_dir / "example3.fits"
    create_fits_image(out_fits2, image_size=(300, 300))

    with pytest.raises(ValueError):
        trim_fits_image(image_path=out_fits2, bounding_box=og_trim.bounding_box)


def test_bounding_box():
    """Create a bounding box around a region of valid non-nan/inf pixels"""
    data = np.zeros((1000, 1000))
    data[10:600, 20:500] = 1
    data[data == 0] = np.nan

    bb = create_bound_box(image_data=data)

    assert isinstance(bb, BoundingBox)
    assert bb.xmin == 10
    assert bb.xmax == 599  # slices upper limit is not inclusive
    assert bb.ymin == 20
    assert bb.ymax == 499  # slices upper limit no inclusive


def test_bounding_box_none():
    """Return None if there are no valid pixels to create a bounding box around"""
    data = np.zeros((1000, 1000)) * np.nan

    bb = _create_bound_box_plane(image_data=data)
    assert bb is None

    bb = create_bound_box(image_data=data)
    assert isinstance(bb, BoundingBox)
    assert bb.xmin == 0
    assert bb.xmin == 0
    assert bb.xmax == 999
    assert bb.ymax == 999


def test_bounding_box_cube():
    """Cube cut bounding boxes."""
    data = np.zeros((3, 1000, 1000))
    data[:, 10:600, 20:500] = 1
    data[data == 0] = np.nan

    with pytest.raises(AssertionError):
        _create_bound_box_plane(image_data=data)

    bb = create_bound_box(image_data=data)
    assert isinstance(bb, BoundingBox)
    assert bb.xmin == 10
    assert bb.xmax == 599
    assert bb.ymin == 20
    assert bb.ymax == 499


def test_bounding_box_cube_different_bounds():
    """Cube cut bounding boxes, where the largest bounding box that
    captures all valid pixels"""
    data = np.zeros((3, 1000, 1000))
    data[0, 10:600, 20:500] = 1
    data[1, 100:200, 600:800] = 1
    data[2, 800:888, 20:500] = 1

    data[data == 0] = np.nan

    with pytest.raises(AssertionError):
        _create_bound_box_plane(image_data=data)

    bb = create_bound_box(image_data=data)
    assert isinstance(bb, BoundingBox)
    assert bb.xmin == 10
    assert bb.xmax == 887
    assert bb.ymin == 20
    assert bb.ymax == 799


def test_bounding_box_with_mask():
    """Create a bounding box where the input is converted to a boolean arrau"""
    data = np.zeros((1000, 1000))
    data[10:600, 20:500] = 1

    bb = create_bound_box(image_data=data, is_masked=True)

    assert isinstance(bb, BoundingBox)
    assert bb.xmin == 10
    assert bb.xmax == 599  # slices upper limit is not inclusive
    assert bb.ymin == 20
    assert bb.ymax == 499  # slices upper limit no inclusive



"""Procedure to calibrate bandpass observation"""

from __future__ import annotations

from argparse import ArgumentParser
from pathlib import Path

import numpy as np
from casacore.tables import table, taql

from flint.calibrate.aocalibrate import AOSolutions, calibrate_apply_ms
from flint.flagging import flag_ms_aoflagger
from flint.logging import logger
from flint.ms import describe_ms, get_field_id_for_field, preprocess_askap_ms
from flint.naming import create_ms_name
from flint.options import MS
from flint.sky_model import KNOWN_1934_FILES, get_1934_model


def plot_solutions(solutions_path: Path, ref_ant: int | None = 0) -> None:
    """Plot solutions for AO-style solutions

    Args:
        solutions_path (Path): Path to the solutions file
        ref_ant (Optional[int], optional): Reference antenna to use. If None is specified there is no division by a reference antenna.  Defaults to 0.
    """
    logger.info(f"Plotting {solutions_path}")

    ao_sols = AOSolutions.load(path=solutions_path)
    _ = ao_sols.plot_solutions(ref_ant=ref_ant)


def flag_bandpass_offset_pointings(ms: MS | Path) -> MS:
    """The typical bandpass style observation in ASKAP will shift each beam
    so that it is centred on the bandpass-calibration object (here B1934-638).
    During each offset position all beams are recording data still. The trick
    here is that all 36 fields are still recorded in the measurement set, and
    are generally of the form:

    >>> ['B1934-638_beam0', 'B1934-638_beam1', 'B1934-638_beam10', 'B1934-638_beam11',
    >>> 'B1934-638_beam12', 'B1934-638_beam13', 'B1934-638_beam14', 'B1934-638_beam15',
    >>> 'B1934-638_beam16', 'B1934-638_beam17', 'B1934-638_beam18', 'B1934-638_beam19',
    >>> 'B1934-638_beam2', 'B1934-638_beam20', 'B1934-638_beam21', 'B1934-638_beam22',
    >>> 'B1934-638_beam23', 'B1934-638_beam24', 'B1934-638_beam25', 'B1934-638_beam26',
    >>> 'B1934-638_beam27', 'B1934-638_beam28', 'B1934-638_beam29', 'B1934-638_beam3',
    >>> 'B1934-638_beam30', 'B1934-638_beam31', 'B1934-638_beam32', 'B1934-638_beam33',
    >>> 'B1934-638_beam34', 'B1934-638_beam35', 'B1934-638_beam4', 'B1934-638_beam5',
    >>> 'B1934-638_beam6', 'B1934-638_beam7', 'B1934-638_beam8', 'B1934-638_beam9']

    This function will attempt to deduce the intended field name for the beam
    in question, and then flag all other fields.

    Args:
        ms (Union[MS, Path]): Path or instance of MS describing the measurement set to flag all other bandpass field.

    Returns:
        MS: A description of the ms
    """
    ms = MS(path=ms) if isinstance(ms, Path) else ms
    ms_summary = describe_ms(ms, verbose=False)

    good_field_name = f"B1934-638_beam{ms_summary.beam}"
    logger.info(f"The B1934-638 field name is {good_field_name}. ")
    logger.info("Will attempt to flag other fields. ")

    with table(f"{ms.path!s}/FIELD", readonly=True, ack=False) as tab:
        # The ID is _position_ of the matching row in the table.
        field_names = tab.getcol("NAME")
        field_idx = np.argwhere([fn == good_field_name for fn in field_names])[0]

        assert len(field_idx) == 1, (
            f"More than one matching field name found. This should not happen. {good_field_name=} {field_names=}"
        )

        field_idx = field_idx[0]
        logger.info(f"{good_field_name} FIELD_ID is {field_idx}")

    with table(f"{ms.path!s}", readonly=False, ack=False) as tab:
        field_idxs = tab.getcol("FIELD_ID")
        field_mask = field_idxs != field_idx
        logger.info(
            f"Found {np.sum(field_mask)} rows not matching FIELD_ID={field_idx}"
        )

        # This is asserting that the stored polarisations are all XX, XY, YX, YY
        flag_row = np.array([True, True, True, True])
        flags = tab.getcol("FLAG")
        flags[field_mask] = flag_row

        # Update the column again
        tab.putcol("FLAG", flags)

    return ms


def extract_correct_bandpass_pointing(
    ms: MS | Path,
    source_name_prefix: str = "B1934-638",
    ms_out_dir: Path | None = None,
) -> MS:
    """The typical bandpass style observation in ASKAP will shift each beam
    so that it is centred on the bandpass-calibration object (here B1934-638).
    During each offset position all beams are recording data still. The trick
    here is that all 36 fields are still recorded in the measurement set, and are
    generally of the form:

    >>> ['B1934-638_beam0', 'B1934-638_beam1', 'B1934-638_beam10', 'B1934-638_beam11',
    >>> 'B1934-638_beam12', 'B1934-638_beam13', 'B1934-638_beam14', 'B1934-638_beam15',
    >>> 'B1934-638_beam16', 'B1934-638_beam17', 'B1934-638_beam18', 'B1934-638_beam19',
    >>> 'B1934-638_beam2', 'B1934-638_beam20', 'B1934-638_beam21', 'B1934-638_beam22',
    >>> 'B1934-638_beam23', 'B1934-638_beam24', 'B1934-638_beam25', 'B1934-638_beam26',
    >>> 'B1934-638_beam27', 'B1934-638_beam28', 'B1934-638_beam29', 'B1934-638_beam3',
    >>> 'B1934-638_beam30', 'B1934-638_beam31', 'B1934-638_beam32', 'B1934-638_beam33',
    >>> 'B1934-638_beam34', 'B1934-638_beam35', 'B1934-638_beam4', 'B1934-638_beam5',
    >>> 'B1934-638_beam6', 'B1934-638_beam7', 'B1934-638_beam8', 'B1934-638_beam9']

    This function will attempt to deduce the intended field name for the beam
    in question, and then create a new measurement set with just these data. It
    internally uses `taql` to create a selection statement:

    >>> with table(ms_path) as tab:
    >>>    sub_ms = taql("select * from $tab where FIELD_ID==field_idx")
    >>>    sub_ms.copy(out_path, deep=True)

    Therefore, some properties from other tables (e.g. FIELDS) may still
    contain references to other fields.

    Args:
        ms (Union[MS, Path]): Path or instance of MS describing the measurement set to flag all other bandpass field.
        source_name_prefix (str, optional): The beginning of the source name stored in the NAME column of the FIELD table. Field names are of the form B1934-638_beam1, where B1934-638 would be the prefix name, and beam is constructed based on the beam among the 36 observing the target source (for example).
        ms_out_dir (Optional[Path], optional): If not None, place the split field measurement sets into this directory. Defaults to None.

    Returns:
        MS: A description of the new measurement set created with the file name ending .beamN.ms.
    """
    ms = MS.cast(ms)
    ms_summary = describe_ms(ms, verbose=False)

    logger.info(f"Checking for unique fields in {ms.path!s} data table.")
    with table(str(ms.path)) as tab:
        fields = np.unique(tab.getcol("FIELD_ID"))
        if len(fields) == 1:
            logger.info(
                f"Only a single field {fields} found. MS likely already split. "
            )

            return ms.with_options(beam=ms_summary.beam)

    good_field_name = f"{source_name_prefix}_beam{ms_summary.beam}"
    field_id = get_field_id_for_field(ms=ms, field_name=good_field_name)

    out_name = create_ms_name(ms_path=ms.path, field=f"{source_name_prefix}")
    out_path = Path("./")
    # Handle writing out to elected output directory.
    # TODO: Move this to a helper utility.
    if ms_out_dir:
        if not ms_out_dir.exists():
            logger.info(f"Will create {ms_out_dir=}")
            try:
                ms_out_dir.mkdir(parents=True)
            except Exception as e:
                logger.warning(f"Exception caught when making {ms_out_dir=}.")
                logger.warning(f"{e}")
                pass

        out_path = ms_out_dir / Path(out_name).name

    logger.info(f"Will create a MS, writing to {out_path}")
    with table(f"{ms.path!s}") as tab:
        field_ms = taql(f"select * from $tab where FIELD_ID=={field_id}")
        field_ms.copy(str(out_path), deep=True)

    return ms.with_options(path=out_path, beam=ms_summary.beam)


def calibrate_bandpass(
    ms_path: Path,
    data_column: str,
    mode: str,
    calibrate_container: Path,
    plot: bool = True,
    aoflagger_container: Path | None = None,
    ms_out_dir: Path | None = None,
) -> MS:
    """Entry point to extract the appropriate field from a bandpass observation,
    run AO-style calibrate, and plot results. In its current form a new measurement
    set will be created container just the appropriate field to calibrate.

    Args:
        ms_path (Path): Path the the measurement set containing bandpass observations of B1934-638
        data_column (str): The column that will be calibrated.
        mode (str): The calibration approach to use. Currently only `calibrate` is supported.
        calibrate_container (Path): The path to the singularity container that holds the appropriate software.
        plot (bool, optional): Whether plotting should be performed. Defaults to True.
        aoflagger_container (Path): The path to the singularity container that holds aoflagger. If this is not `None` that flagging will be performed on the extracted field-specific measurement set.
        ms_out_dir (Optional[Path], optional): If not None, place the split field measurement sets into this directory. Defaults to None.

    Returns:
        MS: The calibrated measurement set with nominated column
    """
    logger.info(f"Will calibrate {ms_path!s}, column {data_column}")

    # TODO: Check to make sure only 1934-638
    model_path: Path = get_1934_model(mode=mode)

    ms = extract_correct_bandpass_pointing(ms=ms_path, ms_out_dir=ms_out_dir)
    describe_ms(ms=ms)

    ms = preprocess_askap_ms(ms)

    if aoflagger_container is not None:
        for i in range(3):
            logger.info("Will run flagger on extracted measurement set. ")
            flag_ms_aoflagger(
                ms=ms.with_options(column="DATA"), container=aoflagger_container
            )

    apply_solutions = calibrate_apply_ms(
        ms_path=ms.path,
        model_path=model_path,
        container=calibrate_container,
        data_column=data_column,
    )

    if plot:
        plot_solutions(solutions_path=apply_solutions.solution_path)

    describe_ms(ms=apply_solutions.ms)

    return ms


def get_parser() -> ArgumentParser:
    parser = ArgumentParser(
        description="Bandpass calibration procedure to calibrate a 1934-638 measurement set. "
    )

    subparser = parser.add_subparsers(
        dest="mode", help="Operation mode of flint_bandpass"
    )
    band_parser = subparser.add_parser("calibrate", help="Perform bandpass calibration")

    supported_models = list(KNOWN_1934_FILES.keys())

    band_parser.add_argument(
        "ms", type=Path, help="Path to the 1934-638 measurement set. "
    )
    band_parser.add_argument(
        "--data-column",
        type=str,
        default="DATA",
        help="Column containing the data to calibrate. ",
    )
    band_parser.add_argument(
        "--mode",
        type=str,
        default="calibrate",
        choices=supported_models,
        help=f"Set the style of the  1934-638 calibration model to use, which depends on the calibration software. available models: {supported_models}. ",
    )
    band_parser.add_argument(
        "--calibrate-container",
        type=Path,
        default=Path("calibrate.sif"),
        help="Path to container that is capable or running and apply calibration solutions for desired mode. ",
    )
    band_parser.add_argument(
        "--aoflagger-container",
        type=Path,
        default=None,
        help="Path to container container aoflagger. If provided guided automated flagging with aoflagger will be performed on the extracted measurement set.  ",
    )
    band_parser.add_argument(
        "--ms-out-dir",
        type=Path,
        default=None,
        help="Location to write the output MSs to. ",
    )

    plot_parser = subparser.add_parser("plot", help="Plot a previous bandpass solution")

    plot_parser.add_argument(
        "solutions", type=Path, help="Path to the AO-style solutions file to plot"
    )
    plot_parser.add_argument(
        "--plot-path",
        type=Path,
        default=None,
        help="Path to write plot to. If None, it is automatically assigned based on solutions name. ",
    )

    return parser


def cli() -> None:
    import logging

    logger.setLevel(logging.INFO)

    parser = get_parser()

    args = parser.parse_args()

    if args.mode == "calibrate":
        calibrate_bandpass(
            ms_path=args.ms,
            data_column=args.data_column,
            mode=args.mode,
            calibrate_container=args.calibrate_container,
            aoflagger_container=args.aoflagger_container,
            ms_out_dir=args.ms_out_dir,
        )
    elif args.mode == "plot":
        plot_solutions(solutions_path=args.solutions)
    else:
        logger.warning("This should not have happened. ")


if __name__ == "__main__":
    cli()



"""Assorted bits and pieces to help with prefect interactions.

Not intended to contain any common task decorated functions
"""

from __future__ import annotations

from prefect import get_run_logger


# This function has been lifted from
# https://gist.github.com/anna-geller/0b9e6ecbde45c355af425cd5b97e303d
# like any good pirate why make when you can take
def enable_loguru_support() -> None:
    """Redirect loguru logging messages to the prefect run logger.
    This function should be called from within a Prefect task or flow before calling any module that uses loguru.
    This function can be safely called multiple times.

    Example Usage:

    >>> from prefect import flow
    >>> from loguru import logger
    >>> from prefect_utils import enable_loguru_support # import this function in your flow from your module
    >>> @flow()
    >>> def myflow():
    >>>     logger.info("This is hidden from the Prefect UI")
    >>>     enable_loguru_support()
    >>>     logger.info("This shows up in the Prefect UI")

    """
    # import here for distributed execution because loguru cannot be pickled.
    from loguru import logger  # pylint: disable=import-outside-toplevel

    run_logger = get_run_logger()
    logger.remove()
    log_format = "{name}:{function}:{line} - {message}"
    logger.add(
        run_logger.debug,
        filter=lambda record: record["level"].name == "DEBUG",
        level="TRACE",
        format=log_format,
    )
    logger.add(
        run_logger.warning,
        filter=lambda record: record["level"].name == "WARNING",
        level="TRACE",
        format=log_format,
    )
    logger.add(
        run_logger.error,
        filter=lambda record: record["level"].name == "ERROR",
        level="TRACE",
        format=log_format,
    )
    logger.add(
        run_logger.critical,
        filter=lambda record: record["level"].name == "CRITICAL",
        level="TRACE",
        format=log_format,
    )
    logger.add(
        run_logger.info,
        filter=lambda record: record["level"].name
        not in ["DEBUG", "WARNING", "ERROR", "CRITICAL"],
        level="TRACE",
        format=log_format,
    )



"""Nothing more than a cmmon place to hold functions
that could be used across many tests"""

from __future__ import annotations

import shutil
from pathlib import Path

import pytest

from flint.utils import get_packaged_resource_path


@pytest.fixture
def ms_example(tmpdir):
    def _ms_example(output_name: str):
        ms_zip = Path(
            get_packaged_resource_path(
                package="flint.data.tests",
                filename="SB39400.RACS_0635-31.beam0.small.ms.zip",
            )
        )
        outpath = Path(tmpdir) / output_name
        if outpath.exists():
            message = f"{outpath=} already exists. Provide unique {output_name=}"
            raise FileExistsError(message)

        shutil.unpack_archive(ms_zip, outpath)

        return Path(outpath) / "SB39400.RACS_0635-31.beam0.small.ms"

    return _ms_example



"""Simple interface into wsclean

Some notes around the file naming.

A certain filenaming scheme is required for FITS files to perform leakage correction
when co-adding them together in the yandasoft linmos application. The stokes field
needs to be encoded as ``.i.``. For example:

    >>> `SB1234.RACS_0123+43.beam01.i.image.fits`

The wsclean formatted output string appends something denoted with ``-``. Within
this code there is and attempt to rename the wsclean outputs to replace the ``-``
with a ``.``.

"""

from __future__ import annotations

import re
from argparse import ArgumentParser
from glob import glob
from numbers import Number
from pathlib import Path
from typing import Any, Collection, NamedTuple

import numpy as np
from astropy.io import fits
from fitscube.combine_fits import combine_fits

from flint.exceptions import (
    AttemptRerunException,
    CleanDivergenceError,
    NamingException,
    NotSupportedError,
)
from flint.logging import logger
from flint.naming import (
    create_image_cube_name,
    create_imaging_name_prefix,
    split_images,
)
from flint.options import (
    MS,
    BaseOptions,
    add_options_to_parser,
    create_options_from_parser,
    options_to_dict,
)
from flint.sclient import run_singularity_command
from flint.utils import (
    get_environment_variable,
    hold_then_move_into,
    remove_files_folders,
)


class ImageSet(BaseOptions):
    """A structure to represent the images and auxiliary products produced by
    wsclean"""

    prefix: str
    """Prefix of the images and other output products. This should correspond to the -name argument from wsclean"""
    image: list[Path]
    """Images produced. """
    psf: list[Path] | None = None
    """References to the PSFs produced by wsclean. """
    dirty: list[Path] | None = None
    """Dirty images. """
    model: list[Path] | None = None
    """Model images.  """
    residual: list[Path] | None = None
    """Residual images."""
    source_list: Path | None = None
    """Path to a source list that accompanies the image data"""


class WSCleanOptions(BaseOptions):
    """A basic container to handle WSClean options. These attributes should
    conform to the same option name in the calling signature of wsclean

    Basic support for environment variables is available. Should a value start
    with `$` it is assumed to be a environment variable, it is will be looked up.
    Some basic attempts to determine if it is a path is made.

    Should the `temp_dir` options be specified then all images will be
    created in this location, and then moved over to the same parent directory
    as the imaged MS. This is done by setting the wsclean `-name` argument.
    """

    abs_mem: int = 100
    """Memory wsclean should try to limit itself to"""
    local_rms_window: int | None = None
    """Size of the window used to estimate rms noise"""
    size: int = 10128
    """Image size, only a single dimension is required. Note that this means images will be squares. """
    local_rms: bool = True
    """Whether a local rms map is computed"""
    force_mask_rounds: int | None = None
    """Round of force masked derivation"""
    auto_mask: float | None = 3.5
    """How deep the construct clean mask is during each cycle"""
    auto_threshold: float | None = 0.5
    """How deep to clean once initial clean threshold reached"""
    threshold: float | None = None
    """Threshold in Jy to stop cleaning"""
    channels_out: int = 4
    """Number of output channels"""
    mgain: float = 0.7
    """Major cycle gain"""
    nmiter: int = 15
    """Maximum number of major cycles to perform"""
    niter: int = 750000
    """Maximum number of minor cycles"""
    multiscale: bool = False
    """Enable multiscale deconvolution"""
    multiscale_scale_bias: float = 0.75
    """Multiscale bias term"""
    multiscale_gain: float | None = None
    """Size of step made in the subminor loop of multi-scale. Default currently 0.2, but shows sign of instability. A value of 0.1 might be more stable."""
    multiscale_scales: tuple[int, ...] | None = None
    """Scales used for multi-scale deconvolution"""
    fit_spectral_pol: int | None = None
    """Number of spectral terms to include during sub-band subtraction"""
    weight: str = "briggs -0.5"
    """Robustness of the weighting used"""
    data_column: str = "CORRECTED_DATA"
    """Which column in the MS to image"""
    scale: str = "2.5asec"
    """Pixel scale size"""
    gridder: str | None = "wgridder"
    """Use the wgridder kernel in wsclean (instead of the default w-stacking method)"""
    nwlayers: int | None = None
    """Number of w-layers to use if the gridder mode is w-stacking"""
    wgridder_accuracy: float | None = None
    """The accuracy requested of the wgridder (should it be used), compared as the RMS error when compred to a DFT"""
    join_channels: bool = True
    """Collapse the sub-band images down to an MFS image when peak-finding"""
    squared_channel_joining: bool = False
    """Use with -join-channels to perform peak finding in the sum of squared values over
    channels, instead of the normal sum. This is useful for imaging QU polarizations
    with non-zero rotation measures, for which the normal sum is insensitive.
    """
    join_polarizations: bool = False
    """Perform deconvolution by searching for peaks in the sum of squares of the polarizations,
    but subtract components from the individual images. Only possible when imaging two or four Stokes
    or linear parameters. Default: off.
    """
    minuv_l: float | None = None
    """The minimum lambda length that the visibility data needs to meet for it to be selected for imaging"""
    minuvw_m: float | None = None
    """A (u,v) selection command, where any baselines shorter than this will be ignored during imaging"""
    maxw: float | None = None
    """A percentage specifying the maximum w-term to be gridded, relative to the max w-term being considered"""
    no_update_model_required: bool = False
    """Will instruct wsclean not to create the MODEL_DATA column"""
    no_small_inversion: bool = False
    """Disables an optimisation of wsclean's w-gridder mode. This might improve accuracy of the w-gridder. """
    beam_fitting_size: float | None = 1.25
    """Use a fitting box the size of <factor> times the theoretical beam size for fitting a Gaussian to the PSF."""
    fits_mask: Path | None = None
    """Path to a FITS file that encodes a cleaning mask"""
    deconvolution_channels: int | None = None
    """The channels out will be averaged down to this many sub-band images during deconvolution"""
    parallel_deconvolution: int | None = None
    """If not none, then this is the number of sub-regions wsclean will attempt to divide and clean"""
    parallel_gridding: int | None = None
    """If not none, then this is the number of channel images that will be gridded in parallel"""
    temp_dir: str | Path | None = None
    """The path to a temporary directory where files will be written. """
    pol: str = "i"
    """The polarisation to be imaged"""
    save_source_list: bool = False
    """Saves the found clean components as a BBS/DP3 text sky model"""
    channel_range: tuple[int, int] | None = None
    """Image a channel range between a lower (inclusive) and upper (exclusive) bound"""
    no_reorder: bool = False
    """If True turn off the reordering of the MS at the beginning of wsclean"""
    flint_no_log_wsclean_output: bool = False
    """If True do not log the wsclean output"""
    no_mf_weighting: bool = False
    """Opposite of -ms-weighting; can be used to turn off MF weighting in -join-channels mode"""


class WSCleanResult(BaseOptions):
    """Simple container for a wsclean command."""

    cmd: str
    """The constructede wsclean command that would be executed."""
    options: WSCleanOptions
    """The set of wslean options used for imaging"""
    ms: MS
    """The measurement sets that have been included in the wsclean command. """
    bind_dirs: tuple[Path, ...]
    """Paths that should be binded to when executing the command"""
    move_hold_directories: tuple[Path, Path]
    """Paths that should be moved to and from when executing the command"""
    image_prefix_str: str | None = None
    """The prefix of the images that will be created"""
    cleanup: bool = True
    """Will clean up the dirty images/psfs/residuals/models when the imaging has completed"""
    image_set: ImageSet | None = None
    """The set of images produced by wsclean"""


def image_set_from_result(wsclean_result: WSCleanResult) -> ImageSet | None:
    return wsclean_result.image_set


def combine_images_to_cube(
    images: list[Path],
    prefix: str,
    mode: str,
    remove_original_images: bool = False,
) -> Path:
    """Combine wsclean subband channel images into a cube. Each collection attribute
    of the input `image_set` will be inspected. The MFS images will be ignored.

    A output file name will be generated based on the  prefix and mode (e.g. `image`, `residual`, `psf`, `dirty`).

    Args:
        image_set (ImageSet): Collection of wsclean image productds
        remove_original_images (bool, optional): If True, images that went into the cube are removed. Defaults to False.

    Returns:
        ImageSet: Updated iamgeset describing the new outputs
    """
    logger.info("Combining subband images into fits cubes")

    output_cube_name = create_image_cube_name(image_prefix=Path(prefix), mode=mode)

    logger.info(f"Combining {len(images)} images. {images=}")
    freqs = combine_fits(file_list=images, out_cube=output_cube_name)
    rotate_cube(output_cube_name)

    # Write out the hdu to preserve the beam table constructed in fitscube
    logger.info(f"Writing {output_cube_name=}")

    output_freqs_name = output_cube_name.with_suffix(".freqs_Hz.txt")
    np.savetxt(output_freqs_name, freqs.to("Hz").value)

    if remove_original_images:
        remove_files_folders(*images)

    return output_cube_name


def merge_image_sets(
    image_sets: list[ImageSet],
) -> ImageSet:
    """Merge the image sets from multiple wsclean results into a single image set.

    Args:
        wsclean_results (List[WSCleanResult]): The results of multiple wsclean imaging runs

    Returns:
        ImageSet: The merged image set
    """
    logger.info(f"Merging {len(image_sets)} image sets")

    prefix = image_sets[0].prefix
    for image_set in image_sets:
        if image_set.prefix != prefix:
            msg = f"Image sets have different prefixes - will use the first prefix {prefix}"
            logger.warning(msg)

    image_set_dict: dict[str, Any] = {}
    for image_set in image_sets:
        for key, value in image_set._asdict().items():
            if key == "prefix":
                continue
            if value is None:
                continue
            if key not in image_set_dict:
                image_set_dict[key] = []
            if isinstance(value, (list, tuple)):
                image_set_dict[key].extend(value)
            else:
                image_set_dict[key].append(value)

    image_set_dict["prefix"] = prefix

    logger.info(f"Merged image set: {image_set_dict=}")

    return ImageSet(**image_set_dict)


def merge_image_sets_from_results(
    wsclean_results: list[WSCleanResult],
) -> ImageSet:
    """Merge the image sets from multiple wsclean results into a single image set.

    Args:
        wsclean_results (List[WSCleanResult]): The results of multiple wsclean imaging runs

    Returns:
        ImageSet: The merged image set
    """
    image_sets = [
        result.image_set for result in wsclean_results if result.image_set is not None
    ]
    return merge_image_sets(image_sets)


def split_image_set(
    image_set: ImageSet,
    by: str = "pol",
    mode: str = "image",
) -> dict[str, list[Path]]:
    """Split an ImageSet by a given field.

    Args:
        image_set (ImageSet): The ImageSet to split
        by (str, optional): Field to split images by. Defaults to "pol".
        mode (str, optional): Type of images to extract from ImageSet. Defaults to "image".

    Raises:
        NamingException: If the field to split by is not found in the ImageSet

    Returns:
        dict[str, list[Path]]: _description_
    """
    logger.info(f"Splitting {image_set=} by {by=} with {mode=}")
    try:
        image_list = getattr(image_set, mode)
    except AttributeError as e:
        msg = f"Failed to extract {mode=} from {image_set=}"
        raise NamingException(msg) from e
    split_list = split_images(images=image_list, by=by)

    return split_list


def split_and_get_image_set(
    image_set: ImageSet,
    get: str,
    by: str = "pol",
    mode: str = "image",
) -> list[Path]:
    """Split an ImageSet by a given field and return the images that match the field of interest.

    Args:
        image_set (ImageSet): The ImageSet to split
        get (str): The field to extract from the split images
        by (str, optional): The field to split by. Defaults to "pol".
        mode (str, optional): The mode to extract from ImageSet. Defaults to "image".

    Raises:
        NamingException: If the field to split by is not found in the ImageSet

    Returns:
        list[Path]: The images that match the field of interest
    """
    split_dict = split_image_set(image_set=image_set, by=by, mode=mode)

    split_list = split_dict.get(get, None)
    if split_list is None:
        raise NamingException(f"Failed to get {get=} from {split_dict=}")

    return split_list


def get_wsclean_output_source_list_path(
    name_path: str | Path, pol: str | None = None
) -> Path:
    """WSClean can produce a text file that describes the components
    that it cleaned, their type, scale and brightness. These are
    placed in a file that is:

    >> {name}.{pol}-sources.txt

    where ``name`` represented the `-name` component. Given
    an input measurement set path or this `-name` value return
    the expected source list text file. ``pol`` is the stokes
    that the source is expected.

    Args:
        name_path (Union[str,Path]): Value of the ``-name`` option. If `str` converted to a ``Path``
        pol (Optional[str], optional): The polarisation to add to the name. If None the -source.txt suffix is simply appended. Defaults to None.

    Returns:
        Path: Path to the source list text file
    """

    # ye not be trusted
    name_path = Path(name_path)
    base_name = name_path.name
    if ".ms" == Path(base_name).suffix:
        base_name = Path(base_name).stem

    logger.info(f"{base_name=} extracted from {name_path=}")

    source_list_name = (
        f"{base_name}.{pol}-sources.txt" if pol else f"{base_name}-sources.txt"
    )
    source_list_path = name_path.parent / source_list_name

    return source_list_path


def _rename_wsclean_title(name_str: str) -> str:
    """Construct and apply a regular expression that aims to identify
    the wsclean appended properties string within a file and replace
    the `-` separator with a `.`.

    A simple replace of all `-` with `.` may not be ideal if the
    character has been used on purpose.

    Args:
        name_str (str): The name that will be extracted and modified

    Returns:
        str: The modified string if a wsclean string was matched, otherwise the input `name-str`
    """
    # Yolo chatGPT
    # The following should replace the .qu with .q or .u whilst removing the -Q and -U
    logger.info(f"Renaming {name_str=} for qu components if necessary")
    name_str = re.sub(
        r"(\.qu)-([^-]+)-?([QU])?(\-(psf|image|dirty|model|residual)\.fits)",
        lambda m: f".{m.group(3).lower() if m.group(3) else 'q'}-{m.group(2)}{m.group(4)}",
        name_str,
    )

    # Note the negative look back - the (?<!...) bit. This is to make
    # sure that the ch0000-0001 field is not captured. For example:
    # SB57516.RACS_0929-81.beam35.round4.i.ch0287-0288-image.fits
    # would be inadvertently picked up and replaced. The [0-9]{4}
    # we are testing for are the channel indicator when --channel-out is used
    search_re = r"(-(i|q|u|v|xx|xy|yx|yy))?(-(MFS|(<!ch[0-9]{4}-[0-9]{4})[0-9]{4}))?(-t[0-9]{5})?-(image|dirty|model|residual|psf)"
    match_re = re.compile(search_re)

    logger.info(f"Searching {name_str=} for wsclean added filename components")
    result = match_re.search(str(name_str))

    if result is None:
        return name_str

    name = name_str.replace(result[0], result[0].replace("-", "."))

    return name


def _rename_wsclean_file(
    input_path: Path,
    rename_file: bool = False,
) -> Path:
    """Rename a file with wsclean appended string information to convert its
    `-` separation markers with `.`. This should handle skipping the field
    name of the target field observed.

    Args:
        input_path (Path): The file path that would be examined and modified
        clean_parts (Union[int, Tuple[int, ...]], optional): Which parts of a split string will be modified. Defaults to -2.
        rename_file (bool, optional): Whether the file should be moved / renamed. Defaults to False.

    Returns:
        Path: Path to the renamed file
    """
    input_path = Path(input_path)
    file_name = Path(input_path.name)
    new_name = _rename_wsclean_title(name_str=str(file_name))

    new_path = input_path.parent / new_name

    if rename_file:
        logger.info(f"Renaming {input_path} to {new_path}")
        input_path.rename(new_path)

    return new_path


def _wsclean_output_callback(line: str) -> None:
    """Call back function used to detect clean divergence"""

    assert isinstance(line, str)

    if "Iteration" in line and "KJy" in line:
        raise CleanDivergenceError(f"Clean divergence detected: {line}")

    # The second commented out line may be related to rerunning a wsclean task with a
    # existing set of temporary files in the wsclean context. Look into this.
    # TODO: Look at behaviour of wsclean should it be executed in a location where tempoary
    # files from a previous execution exist
    # temp_error_lines = ("Error opening temporary data file", "Input/output error")

    temp_error_lines = ("Input/output error",)
    if any([temp_error_line in line for temp_error_line in temp_error_lines]):
        logger.info(f"Detected input/output error in {line}")
        from time import sleep

        sleep(2)

        raise AttemptRerunException


# TODO: Update this function to also add int the source list
def get_wsclean_output_names(  #
    prefix: str,
    subbands: int | None = None,
    pols: str | tuple[str] | tuple[str, ...] | None = None,
    verify_exists: bool = False,
    include_mfs: bool = True,
    output_types: str | Collection[str] = (
        "image",
        "dirty",
        "residual",
        "model",
        "psf",
    ),
    check_exists_when_adding: bool = False,
) -> ImageSet:
    """Attempt to generate the file names and paths that would be
    created by an imaging run of wsclean. This is done using a the
    known (expected) naming format of modern wsclean versions.

    Checks can be made to ensure a file exists before adding it
    to the output ``ImageSet``. This might be important as some
    wsclean image products might be deleted in order to preserve
    disk space.

    Args:
        prefix (str): The prefix of the imaging run (akin to -name option in wsclean call)
        subbands (int): Number of subbands that were imaged
        pols (Optional[Union[str,Tuple[str]]], optional): The polarisation of the image. If None are provided then this is not used. Multiple polarisation may be supplied. If multiple pols are given in an iterable, each will be produced. Defaults to None.
        verify_exists (bool, optional): Ensures that each generated path corresponds to an actual file. Defaults to False.
        include_mfs (bool, optional): Include the MFS images produced by wsclean. Defaults to True.
        output_types (Union[str,Collection[str]]): Include files of this type, including image, dirty, residual, model, psf. Defaults to  ('image','dirty','residual','model', 'psf').
        check_exists_when_adding (bool, optional): Only add a generated file if it exists. Although related to the ``verify_exists``, when this optional is ``True`` files will silently be ignored if they are not found. if this and ``verify_exists`` are both ``True`` then no errors on missing files will be raised. Defaults to False.

    Raises:
        FileExistsError: Raised when a file does not exist and verify_exists is True.

    Returns:
        ImageSet: The file paths that wsclean should create/has created.
    """
    logger.info(f"Finding wsclean outputs, {prefix=}")
    # TODO: Use a regular expression for this
    subband_strs: list[str | None] = [
        None,
    ]
    if subbands and subbands > 1:
        subband_strs = [f"{subband:04}" for subband in range(subbands)]
        if include_mfs:
            subband_strs.append("MFS")

    in_pols: tuple[None, ...] | tuple[str, ...]
    if pols is None:
        in_pols = (None,)
    elif isinstance(pols, str):
        in_pols = (pols,)
    else:
        in_pols = pols

    if isinstance(output_types, str):
        output_types = (output_types,)

    images: dict[str, list[Path]] = {}
    for image_type in ("image", "dirty", "model", "residual"):
        if image_type not in output_types:
            continue

        paths: list[Path] = []
        for pol in in_pols:
            for subband_str in subband_strs:
                components = [prefix]
                if subband_str:
                    components.append(subband_str)
                if pol:
                    components.append(pol)
                components.append(image_type)
                path_str = "-".join(components) + ".fits"

                if check_exists_when_adding and not Path(path_str).exists():
                    logger.debug(f"{path_str} does not existing. Not adding. ")
                    continue

                paths.append(Path(path_str))

        images[image_type] = paths

    # The PSF is the same for all stokes
    if "psf" in output_types:
        psf_images = [
            (
                Path(f"{prefix}-{subband_str}-psf.fits")
                if subband_str
                else Path(f"{prefix}-psf.fits")
            )
            for subband_str in subband_strs
        ]
        # Filter out files if they do not exists
        if check_exists_when_adding:
            psf_images = [psf_image for psf_image in psf_images if psf_image.exists()]

        images["psf"] = psf_images

    if verify_exists:
        paths_no_exists: list[Path] = []
        for _, check_paths in images.items():
            paths_no_exists += [path for path in check_paths if not path.exists()]
        if len(paths_no_exists) > 0:
            raise FileExistsError(
                f"The following {len(paths_no_exists)} files do not exist: {paths_no_exists}"
            )

    return ImageSet(prefix=prefix, **images)


def delete_wsclean_outputs(
    prefix: str,
    output_type: str = "image",
    ignore_mfs: bool = True,
    no_subbands: bool = False,
) -> list[Path]:
    """Attempt to remove elected wsclean output files

    If ``no_subbands`` is True (as in ``channels_out`` is 1) then nothing is deleted.

    Args:
        prefix (str): The prefix of the files to remove. This would correspond to the -name of wsclean.
        output_type (str, optional): What type of wsclean output to try to remove. Defaults to 'image'.
        ignore_mfs (bool, optional): If True, do not remove MFS outputs (attempt to, at least). Defaults to True.
        no_subbands (bool, Optional): If True, nothing is deleted. Defaults to False.

    Returns:
        List[Path]: The paths that were removed (or at least attempted to be removed)/
    """
    # TODO: This glob needs to be replaced with something more explicit
    paths = [Path(p) for p in glob(f"{prefix}-*{output_type}.fits")]
    logger.info(f"Found {len(paths)} matching {prefix=} and {output_type=}.")
    rm_paths: list[Path] = []

    for path in paths:
        if no_subbands:
            continue
        if ignore_mfs and "-MFS-" in str(path.name):
            logger.info(f"{path=} appears to be an MFS product, not removing. ")
            continue
        if path.exists():
            logger.warning(f"Removing {path}.")
            try:
                rm_paths.append(path)
                path.unlink()
            except Exception as e:
                logger.critical(f"Removing {path} failed: {e}")

    return rm_paths


# TODO: This should be expanded to denote levels of things to delete
# TODO: Need to create a regex based mode, and better handlijng of -MFS-,
# which is only created when -join-channels is used
def wsclean_cleanup_files(
    prefix: str | Path,
    output_types: tuple[str, ...] | None = ("dirty", "psf", "model", "residual"),
    single_channel: bool = False,
) -> tuple[Path, ...]:
    """Clean up (i.e. delete) files from wsclean.

    Args:
        prefix (Union[str, Path]): The prefix used to search for files. This is generally the -name
        output_types (Optional[Tuple[str]], optional): Which type of output wsclean products to delete. Defaults to ("dirty", "psf", "model", "residual").
        single_channel (bool, optional): Whether there is the subband part of the wsclean file names to consider. Defaults to False.

    Returns:
        Tuple[Path, ...]: Set of files that were deleted
    """
    rm_files = []
    logger.info(f"Removing wsclean files with {prefix=} {output_types=}")

    if output_types is not None:
        for output_type in output_types:
            rm_files += delete_wsclean_outputs(
                prefix=str(prefix),
                output_type=output_type,
                no_subbands=single_channel and output_type == "image",
            )

    return tuple(rm_files)


def create_wsclean_name_argument(wsclean_options: WSCleanOptions, ms: MS) -> Path:
    """Create the value that will be provided to wsclean -name argument. This has
    to be generated. Among things to consider is the desired output directory of imaging
    files. This by default will be alongside the measurement set. If a `temp_dir`
    has been specified then output files will be written here.

    Args:
        wsclean_options (WSCleanOptions): Set of wsclean options to consider
        ms (MS): The measurement set to be imaged

    Returns:
        Path: Value of the -name argument to provide to wsclean
    """
    wsclean_options_dict = wsclean_options._asdict()

    # Prepare the name for the output wsclean command
    # Construct the name property of the string
    pol = wsclean_options.pol
    channel_range = wsclean_options.channel_range
    name_prefix_str = create_imaging_name_prefix(
        ms_path=ms.path, pol=pol, channel_range=channel_range
    )

    # Now resolve the directory part
    name_dir: Path | str | None = ms.path.parent
    temp_dir = wsclean_options_dict.get("temp_dir", None)
    if temp_dir:
        # Resolve if environment variable
        name_dir = (
            get_environment_variable(variable=temp_dir)
            if isinstance(temp_dir, str) and temp_dir[0] == "$"
            else Path(temp_dir)
        )

    assert name_dir is not None, f"{name_dir=} is None, which is bad"

    name_argument_path = Path(name_dir) / name_prefix_str
    logger.info(f"Constructed -name {name_argument_path}")

    return name_argument_path


class ResolvedCLIResult(NamedTuple):
    """Mapping results to provide to wsclean"""

    cmd: str | None = None
    """The argument value pair to place on the CLI. """
    unknown: Any | None = None
    """Unknown options that could not be converted"""
    bindpath: Path | None = None
    """A path to bind to when called within a container"""
    ignore: bool = False
    """Ignore this CLIResult if True"""


def _resolve_wsclean_key_value_to_cli_str(key: str, value: Any) -> ResolvedCLIResult:
    """An internal function intended to map a key-value pair to
    the appropriate form to pass to a CLI call into wsclean.

    Args:
        key (str): The wsclean argument name to consider. Underscores will be converted to hyphens, as expected by wsclean
        value (Any): The value of the argument that should be converted to the appropriately formatted string

    Returns:
        ResolvedCLIResult: Converted CLI output, including paths to bind to and unknown conversions
    """

    # Some wsclean options, if multiple values are provided, might need
    # to be join as a csv list. Others might want to be dumped in. Just
    # attempting to future proof (arguably needlessly).
    options_to_comma_join = "multiscale-scales"
    bind_dir_options = ("temp-dir",)

    logger.debug(f"{key=} {value=} {type(value)=}")

    value = (
        get_environment_variable(variable=value)
        if isinstance(value, str) and value[0] == "$"
        else value
    )

    cmd = None
    unknown = None
    bind_dir_path = None

    original_key = key
    key = key.replace("_", "-")

    if original_key.startswith("flint_"):
        # No need to do anything more
        return ResolvedCLIResult(ignore=True)
    elif key == "size":
        cmd = f"-size {value} {value}"
    elif isinstance(value, bool):
        if value:
            cmd = f"-{key}"
    elif isinstance(value, (str, Number)):
        cmd = f"-{key} {value}"
    elif isinstance(value, (list, tuple)):
        value = list(map(str, value))
        value_str = ",".join(value) if key in options_to_comma_join else " ".join(value)
        cmd = f"-{key} {value_str}"
    elif isinstance(value, Path):
        value_str = str(value)
        cmd = f"-{key} {value_str}"
    elif value is None:
        logger.debug(
            f"{key} option set to {value}. Not sure what this means. Ignoring. "
        )
    else:
        unknown = (original_key, value)

    if key in bind_dir_options and isinstance(value, (str, Path)):
        bind_dir_path = Path(value)

    return ResolvedCLIResult(cmd=cmd, unknown=unknown, bindpath=bind_dir_path)


def create_wsclean_cmd(
    ms: MS,
    wsclean_options: WSCleanOptions,
) -> WSCleanResult:
    """Create a wsclean command from a WSCleanOptions container

    For the most part these are one-to-one mappings to the wsclean CLI with the
    exceptions being:
    #. the `-name` argument will be generated and supplied to the CLI string and will default to the parent directory and name of the supplied measurement set
    #. If `wsclean_options.temp_dir` is specified this directory is used in place of the measurement sets parent directory

    If `container` is supplied to immediately execute this command then the
    output wsclean image products will be moved from the `temp-dir` to the
    same directory as the measurement set.

    Args:
        ms (MS): The measurement set to be imaged
        wsclean_options (WSCleanOptions): WSClean options to image with
        container (Optional[Path], optional): If a path to a container is provided the command is executed immediately. Defaults to None.

    Raises:
        ValueError: Raised when a option has not been successfully processed

    Returns:
        WSCleanResult: The wsclean command to run
    """
    # TODO: This is very very smelly. I think removing the `.name` from WSCleanOptions
    # to start with, build that as an explicit testable function, and pass/return the name
    # argument alongside the prefix in the WSCleanCMD. Also need to rename that, its a horrible
    # name for a variable and ship

    # Some options should also extend the singularity bind directories
    bind_dir_paths = []

    name_argument_path = create_wsclean_name_argument(
        wsclean_options=wsclean_options, ms=ms
    )
    move_directory = ms.path.parent
    hold_directory: Path | None = Path(name_argument_path).parent

    wsclean_options_dict = wsclean_options._asdict()

    unknowns: list[tuple[Any, Any]] = []
    logger.info("Creating wsclean command.")

    cli_results = list(
        map(
            _resolve_wsclean_key_value_to_cli_str,
            wsclean_options_dict.keys(),
            wsclean_options_dict.values(),
        )
    )

    # Ignore any CLIResult if it has been explicitly instructed to
    cli_results = [cli_result for cli_result in cli_results if not cli_result.ignore]

    cmds = [cli_result.cmd for cli_result in cli_results if cli_result.cmd]
    unknowns = [cli_result.unknown for cli_result in cli_results if cli_result.unknown]
    bind_dir_paths += [
        cli_result.bindpath for cli_result in cli_results if cli_result.bindpath
    ]

    if len(unknowns) > 0:
        msg = ", ".join([f"{t[0]} {t[1]}" for t in unknowns])
        raise ValueError(f"Unknown wsclean option types: {msg}")

    cmds += [f"-name {name_argument_path!s}"]
    cmds += [f"{ms.path!s} "]

    bind_dir_paths.append(ms.path.parent)

    cmd = "wsclean " + " ".join(cmds)

    logger.info(f"Constructed wsclean command: {cmd=}")
    logger.info("Setting default model data column to 'MODEL_DATA'")

    return WSCleanResult(
        cmd=cmd,
        options=wsclean_options,
        ms=ms.with_options(model_column="MODEL_DATA"),
        bind_dirs=tuple(bind_dir_paths),
        move_hold_directories=(move_directory, hold_directory),
        image_prefix_str=str(name_argument_path),
    )


def rotate_cube(output_cube_name) -> None:
    logger.info(f"Rotating {output_cube_name=}")
    # This changes the output cube to a shape of (chan, pol, dec, ra)
    # which is what yandasoft linmos tasks like
    with fits.open(output_cube_name, mode="update", memmap=True) as hdulist:
        hdu = hdulist[0]
        new_header = hdu.header
        data_cube = hdu.data

        tmp_header = new_header.copy()
        # Need to swap NAXIS 3 and 4 to make LINMOS happy - booo
        for a, b in ((3, 4), (4, 3)):
            new_header[f"CTYPE{a}"] = tmp_header[f"CTYPE{b}"]
            new_header[f"CRPIX{a}"] = tmp_header[f"CRPIX{b}"]
            new_header[f"CRVAL{a}"] = tmp_header[f"CRVAL{b}"]
            new_header[f"CDELT{a}"] = tmp_header[f"CDELT{b}"]
            new_header[f"CUNIT{a}"] = tmp_header[f"CUNIT{b}"]

        # Cube is currently STOKES, FREQ, RA, DEC - needs to be FREQ, STOKES, RA, DEC
        data_cube = np.moveaxis(data_cube, 1, 0)
        hdu.data = data_cube
        hdu.header = new_header
        hdulist.flush()


def combine_image_set_to_cube(
    image_set: ImageSet,
    remove_original_images: bool = False,
) -> ImageSet:
    """Combine wsclean subband channel images into a cube. Each collection attribute
    of the input `image_set` will be inspected. The MFS images will be ignored.

    A output file name will be generated based on the  prefix and mode (e.g. `image`, `residual`, `psf`, `dirty`).

    Args:
        image_set (ImageSet): Collection of wsclean image productds
        remove_original_images (bool, optional): If True, images that went into the cube are removed. Defaults to False.

    Returns:
        ImageSet: Updated iamgeset describing the new outputs
    """
    logger.info("Combining subband image products into fits cubes")

    if not isinstance(image_set, ImageSet):
        raise TypeError(
            f"Input image_set of type {type(image_set)}, expect {type(ImageSet)}"
        )

    image_set_dict = options_to_dict(input_options=image_set)

    for mode in ("image", "residual", "dirty", "model", "psf"):
        if image_set_dict[mode] is None or not isinstance(
            image_set_dict[mode], (list, tuple)
        ):
            logger.info(f"{mode=} is None or not appropriately formed. Skipping. ")
            continue

        subband_images = [
            image for image in image_set_dict[mode] if "-MFS-" not in str(image)
        ]
        if len(subband_images) <= 1:
            logger.info(f"Not enough subband images for {mode=}, not creating a cube")
            continue

        output_cube_name = create_image_cube_name(
            image_prefix=Path(image_set.prefix), mode=mode
        )

        logger.info(f"Combining {len(subband_images)} images. {subband_images=}")
        freqs = combine_fits(file_list=subband_images, out_cube=output_cube_name)

        rotate_cube(output_cube_name)

        # Write out the hdu to preserve the beam table constructed in fitscube
        logger.info(f"Writing {output_cube_name=}")

        output_freqs_name = Path(output_cube_name).with_suffix(".freqs_Hz.txt")
        np.savetxt(output_freqs_name, freqs.to("Hz").value)

        image_set_dict[mode] = [Path(output_cube_name)] + [
            image for image in image_set_dict[mode] if image not in subband_images
        ]

        if remove_original_images:
            remove_files_folders(*subband_images)

    return ImageSet(**image_set_dict)


def rename_wsclean_prefix_in_image_set(input_image_set: ImageSet) -> ImageSet:
    """Given an input image_set, rename the files contained in it to
    remove the `-` separator that wsclean uses and replace it with `.`.

    Files will be renamed on disk appropriately.

    Args:
        input_image_set (ImageSet): The collection of output wsclean products

    Returns:
        ImageSet: The updated image_set after replacing the separator and renaming files
    """

    input_args = options_to_dict(input_options=input_image_set)

    check_keys = ("prefix", "image", "residual", "model", "dirty")

    output_args: dict[str, Any] = {}

    for key, value in input_args.items():
        if key == "prefix":
            output_args[key] = _rename_wsclean_title(name_str=value)
        elif key in check_keys and value is not None:
            output_args[key] = [
                _rename_wsclean_file(input_path=fits_path, rename_file=True)
                for fits_path in value
            ]
        else:
            output_args[key] = value

    output_image_set = ImageSet(**output_args)

    return output_image_set


def _make_pols(pol_str: str) -> tuple[str, ...] | None:
    """Create a tuple of polarisations from a polarisation string

    Args:
        pol_str (str): Polarisation string to convert

    Returns:
        tuple[str, ...] | None: Tuple of polarisations
    """
    if any(pol in pol_str.lower() for pol in ("x", "y", "r", "l")):
        msg = f"Found instrumental polarisation in {pol_str=}. This is not supported."
        raise NotSupportedError(msg)

    pols: tuple[str, ...] | None = None
    if len(pol_str) > 1:
        pols = tuple(p.upper() for p in "".join(pol_str.split(",")))
    return pols


def run_wsclean_imager(
    wsclean_result: WSCleanResult,
    container: Path,
    make_cube_from_subbands: bool = True,
) -> ImageSet:
    """Run a provided wsclean command. Optionally will clean up files,
    including the dirty beams, psfs and other assorted things.

    An `ImageSet` is constructed that attempts to capture the output wsclean image products. If `image_prefix_str`
    is specified the image set will be created by (ordered by preference):
    #. Adding the `image_prefix_str` to the `move_directory`
    #. Guessing it from the path of the measurement set from `wsclean_result.ms.path`

    Args:
        wsclean_result (WSCleanResult): The command to run, and other properties (cleanup.)
        container (Path): Path to the container with wsclean available in it
        bind_dirs (Optional[Tuple[Path, ...]], optional): Additional directories to include when binding to the wsclean container. Defaults to None.
        move_hold_directories (Optional[Tuple[Path,Optional[Path]]], optional): The `move_directory` and `hold_directory` passed to the temporary context manager. If None no `hold_then_move_into` manager is used. Defaults to None.
        make_cube_from_subbands (bool, optional): Form a single FITS cube from the set of sub-band images wsclean produces. Defaults to False.
        image_prefix_str (Optional[str], optional): The name used to search for wsclean outputs. If None, it is guessed from the name and location of the MS. Defaults to None.

    Returns:
        ImageSet: The executed wsclean output products.
    """

    ms = wsclean_result.ms
    single_channel = wsclean_result.options.channels_out == 1
    wsclean_cleanup = wsclean_result.cleanup
    bind_dirs = wsclean_result.bind_dirs
    move_hold_directories = wsclean_result.move_hold_directories
    image_prefix_str = wsclean_result.image_prefix_str

    sclient_bind_dirs = [Path(ms.path).parent.absolute()]
    if bind_dirs:
        sclient_bind_dirs = sclient_bind_dirs + list(bind_dirs)

    prefix = image_prefix_str if image_prefix_str else None
    if prefix is None:
        prefix = str(ms.path.parent / ms.path.name)
        logger.warning(f"Setting prefix to {prefix}. Likely this is not correct. ")

    if move_hold_directories:
        with hold_then_move_into(
            move_directory=move_hold_directories[0],
            hold_directory=move_hold_directories[1],
        ) as directory:
            sclient_bind_dirs.append(directory)
            run_singularity_command(
                image=container,
                command=wsclean_result.cmd,
                bind_dirs=sclient_bind_dirs,
                stream_callback_func=_wsclean_output_callback,
                ignore_logging_output=wsclean_result.options.flint_no_log_wsclean_output,
            )
            if wsclean_cleanup:
                rm_files = wsclean_cleanup_files(
                    prefix=prefix, single_channel=single_channel
                )
                logger.info(f"Removed {len(rm_files)} files")
                # No need to attempt to clean up again once files have been moved

                wsclean_cleanup = False
            # Update the prefix based on where the files will be moved to
            prefix = (
                f"{move_hold_directories[0] / Path(prefix).name!s}"
                if image_prefix_str
                else None
            )
    else:
        run_singularity_command(
            image=container,
            command=wsclean_result.cmd,
            bind_dirs=sclient_bind_dirs,
            stream_callback_func=_wsclean_output_callback,
            ignore_logging_output=wsclean_result.options.flint_no_log_wsclean_output,
        )

    # prefix should be set at this point
    assert prefix is not None, f"{prefix=}, which should not happen"

    if wsclean_cleanup:
        logger.info("Will clean up files created by wsclean. ")
        rm_files = wsclean_cleanup_files(prefix=prefix, single_channel=single_channel)

    pols = _make_pols(pol_str=wsclean_result.options.pol)

    image_set = get_wsclean_output_names(
        prefix=prefix,
        subbands=wsclean_result.options.channels_out,
        pols=pols,
        verify_exists=True,
        output_types=("image", "residual"),
        check_exists_when_adding=True,
    )

    if wsclean_result.options.save_source_list:
        logger.info("Attaching the wsclean clean components SEDs")
        source_list_path = get_wsclean_output_source_list_path(
            name_path=prefix, pol=None
        )
        assert source_list_path.exists(), f"{source_list_path=} does not exist"
        image_set = image_set.with_options(source_list=source_list_path)

    if make_cube_from_subbands:
        image_set = combine_image_set_to_cube(
            image_set=image_set, remove_original_images=True
        )

    image_set = rename_wsclean_prefix_in_image_set(input_image_set=image_set)

    logger.info(f"Constructed {image_set=}")

    return image_set


def wsclean_imager(
    ms: Path | MS,
    wsclean_container: Path,
    update_wsclean_options: dict[str, Any] | None = None,
    make_cube_from_subbands: bool = True,
) -> WSCleanResult:
    """Create and run a wsclean imager command against a measurement set.

    Args:
        ms (Union[Path,MS]): Path to the measurement set that will be imaged
        wsclean_container (Path): Path to the container with wsclean installed
        update_wsclean_options (Optional[Dict[str, Any]], optional): Additional options to update the generated WscleanOptions with. Keys should be attributes of WscleanOptions. Defaults to None.

    Returns:
        WSCleanResult: _description_
    """

    # TODO: This should be expanded to support multiple measurement sets
    ms = MS.cast(ms)

    wsclean_options = WSCleanOptions()
    if update_wsclean_options:
        logger.info("Updatting wsclean options with user-provided items. ")
        wsclean_options = wsclean_options.with_options(**update_wsclean_options)

    assert ms.column is not None, "A MS column needs to be elected for imaging. "
    wsclean_options = wsclean_options.with_options(data_column=ms.column)
    wsclean_result = create_wsclean_cmd(
        ms=ms,
        wsclean_options=wsclean_options,
    )
    image_set = run_wsclean_imager(
        wsclean_result=wsclean_result,
        container=wsclean_container,
        make_cube_from_subbands=make_cube_from_subbands,
    )

    return wsclean_result.with_options(image_set=image_set)


def get_parser() -> ArgumentParser:
    parser = ArgumentParser(description="Routines related to wsclean")

    subparser = parser.add_subparsers(dest="mode")

    wsclean_parser = subparser.add_parser(
        "image", help="Attempt to run a wsclean command. "
    )
    wsclean_parser.add_argument(
        "ms", type=Path, help="Path to a measurement set to image"
    )
    wsclean_parser.add_argument(
        "-v", "--verbose", action="store_true", help="Extra output logging."
    )
    wsclean_parser.add_argument(
        "--wsclean-container",
        type=Path,
        default=None,
        help="Path to a singularity container with wsclean installed. ",
    )
    wsclean_parser = add_options_to_parser(
        parser=wsclean_parser, options_class=WSCleanOptions
    )

    return parser


def cli() -> None:
    parser = get_parser()

    args = parser.parse_args()

    if args.mode == "image":
        if args.verbose:
            import logging

            logger.setLevel(logging.DEBUG)

        ms = MS(path=args.ms, column=args.data_column)
        wsclean_options: WSCleanOptions = create_options_from_parser(
            parser_namespace=args,
            options_class=WSCleanOptions,  # type: ignore
        )
        wsclean_imager(
            ms=ms,
            wsclean_container=args.wsclean_container,
            update_wsclean_options=wsclean_options._asdict(),
        )


if __name__ == "__main__":
    cli()



from __future__ import annotations

import shutil
from pathlib import Path

import numpy as np
import pytest
from astropy.coordinates import EarthLocation, Latitude, Longitude
from astropy.time import Time

from flint.imager.wsclean import ImageSet
from flint.ms import get_telescope_location_from_ms, get_times_from_ms
from flint.source_finding.aegean import AegeanOutputs
from flint.summary import (
    FieldSummary,
    add_rms_information,
    create_beam_summary,
    create_field_summary,
    update_field_summary,
)
from flint.utils import get_packaged_resource_path
from flint.validation import make_psf_table


@pytest.fixture
def aegean_outputs_example():
    rms = Path(
        get_packaged_resource_path(
            package="flint.data.tests",
            filename="SB39400.RACS_0635-31.beam0-MFS-subimage_rms.fits",
        )
    )

    comp = Path(
        get_packaged_resource_path(
            package="flint.data.tests",
            filename="SB38959.RACS_1357-18.noselfcal.linmos_comp.fits",
        )
    )

    aegean_outputs = AegeanOutputs(
        bkg=rms, rms=rms, comp=comp, beam_shape=(1.0, 1.0, 1.0), image=rms
    )

    return aegean_outputs


@pytest.fixture
def ms_example(tmpdir):
    ms_zip = Path(
        get_packaged_resource_path(
            package="flint.data.tests",
            filename="SB39400.RACS_0635-31.beam0.small.ms.zip",
        )
    )
    outpath = Path(tmpdir) / "39400"

    shutil.unpack_archive(ms_zip, outpath)

    ms_path = Path(outpath) / "SB39400.RACS_0635-31.beam0.small.ms"

    return ms_path


def test_create_beam_summary(ms_example, aegean_outputs_example):
    beam_summary = create_beam_summary(ms=ms_example, components=aegean_outputs_example)

    assert beam_summary.ms_summary.path == ms_example


def test_create_field_summary_beam_summary(ms_example, aegean_outputs_example):
    cal_sbid_path = Path("/scratch3/gal16b/split/39433/SB39433.1934-638.beam0.ms")
    mss = [ms_example for _ in range(36)]

    image_set = ImageSet(prefix="Example", image=[aegean_outputs_example.rms])

    beam_summaries = [
        create_beam_summary(
            ms=ms_example, image_set=image_set, components=aegean_outputs_example
        )
        for _ in range(36)
    ]

    field_summary = create_field_summary(
        mss=mss, cal_sbid_path=cal_sbid_path, beam_summaries=beam_summaries
    )

    assert len(field_summary.beam_summaries) == 36


def test_create_field_summary_beam_summary_nocalid(ms_example, aegean_outputs_example):
    """Make sure the summary will work when an empty of non-existent calid is passed in,
    which may happen for MSs downloaded through casda"""
    mss = [ms_example for _ in range(36)]

    image_set = ImageSet(prefix="Example", image=[aegean_outputs_example.rms])

    beam_summaries = [
        create_beam_summary(
            ms=ms_example, image_set=image_set, components=aegean_outputs_example
        )
        for _ in range(36)
    ]

    for item in (None, "None", "none"):
        field_summary = create_field_summary(
            mss=mss, cal_sbid_path=item, beam_summaries=beam_summaries
        )
        assert field_summary.beam_summaries is not None
        assert len(field_summary.beam_summaries) == 36


def test_field_summary_beam_summary_make_psf(
    ms_example, aegean_outputs_example, tmpdir
):
    cal_sbid_path = Path("/scratch3/gal16b/split/39433/SB39433.1934-638.beam0.ms")
    mss = [ms_example for _ in range(36)]
    image_set = ImageSet(prefix="Example", image=[aegean_outputs_example.rms])

    beam_summaries = [
        create_beam_summary(
            ms=ms_example, image_set=image_set, components=aegean_outputs_example
        )
        for _ in range(36)
    ]

    field_summary = create_field_summary(
        mss=mss, cal_sbid_path=cal_sbid_path, beam_summaries=beam_summaries
    )

    make_psf_table(field_summary=field_summary, output_path=Path(tmpdir))


def test_field_summary(ms_example):
    cal_sbid_path = Path("/scratch3/gal16b/split/39433/SB39433.1934-638.beam0.ms")
    mss = [ms_example for _ in range(36)]

    field_summary = create_field_summary(mss=mss, cal_sbid_path=cal_sbid_path)

    assert isinstance(field_summary, FieldSummary)
    assert field_summary.sbid == "39400"
    assert field_summary.cal_sbid == "39433"

    field_summary = create_field_summary(mss=mss)

    assert isinstance(field_summary, FieldSummary)
    assert field_summary.sbid == "39400"
    assert field_summary.cal_sbid == "None"
    assert isinstance(field_summary.location, EarthLocation)
    assert field_summary.integration_time == 19.90655994415036
    assert isinstance(field_summary.ms_times, Time)
    assert len(field_summary.ms_times) == 2


def test_field_summary_with_mss(ms_example, aegean_outputs_example):
    cal_sbid_path = Path("/scratch3/gal16b/split/39433/SB39433.1934-638.beam0.ms")

    mss = [ms_example for _ in range(36)]

    field_summary = create_field_summary(
        cal_sbid_path=cal_sbid_path,
        aegean_outputs=aegean_outputs_example,
        mss=mss,
    )

    # This si the phase direction, in degrees, of the one MS
    # this pirate is sneakily repeating
    centre = field_summary.centre
    assert np.isclose(centre.ra.deg, 98.211959)  # type: ignore
    assert np.isclose(centre.dec.deg, -30.86099889)  # type: ignore

    assert isinstance(field_summary.hour_angles, Longitude)
    assert isinstance(field_summary.elevations, Latitude)


def test_field_summary_rms_info(ms_example, aegean_outputs_example):
    cal_sbid_path = Path("/scratch3/gal16b/split/39433/SB39433.1934-638.beam0.ms")

    mss = [ms_example for _ in range(36)]

    field_summary = create_field_summary(mss=mss, cal_sbid_path=cal_sbid_path)

    assert isinstance(field_summary, FieldSummary)
    assert field_summary.sbid == "39400"
    assert field_summary.cal_sbid == "39433"
    assert isinstance(field_summary.location, EarthLocation)
    assert field_summary.integration_time == 19.90655994415036
    assert isinstance(field_summary.ms_times, Time)

    field_summary = add_rms_information(
        field_summary=field_summary, aegean_outputs=aegean_outputs_example
    )
    assert field_summary.no_components == 7225


def test_field_summary_rms_info_creation(ms_example, aegean_outputs_example):
    cal_sbid_path = Path("/scratch3/gal16b/split/39433/SB39433.1934-638.beam0.ms")

    mss = [ms_example for _ in range(36)]

    field_summary = create_field_summary(
        mss=mss, cal_sbid_path=cal_sbid_path, aegean_outputs=aegean_outputs_example
    )

    assert isinstance(field_summary, FieldSummary)
    assert field_summary.sbid == "39400"
    assert field_summary.cal_sbid == "39433"
    assert isinstance(field_summary.location, EarthLocation)
    assert field_summary.integration_time == 19.90655994415036
    assert isinstance(field_summary.ms_times, Time)


def test_field_summary_update(ms_example, aegean_outputs_example):
    cal_sbid_path = Path("/scratch3/gal16b/split/39433/SB39433.1934-638.beam0.ms")

    mss = [ms_example for _ in range(36)]

    field_summary = create_field_summary(mss=mss, cal_sbid_path=cal_sbid_path)

    assert isinstance(field_summary, FieldSummary)
    assert field_summary.sbid == "39400"
    assert field_summary.cal_sbid == "39433"
    assert isinstance(field_summary.location, EarthLocation)
    assert field_summary.integration_time == 19.90655994415036
    assert isinstance(field_summary.ms_times, Time)
    assert field_summary.round is None

    field_summary = update_field_summary(
        field_summary=field_summary, aegean_outputs=aegean_outputs_example, round=2
    )
    assert field_summary.no_components == 7225
    assert field_summary.round == 2


def test_ms_example_times(ms_example):
    times = get_times_from_ms(ms=ms_example)

    assert isinstance(times, Time)
    assert len(times) == 1998


def test_ms_example_telescope(ms_example):
    telescope = get_telescope_location_from_ms(ms=ms_example)

    assert isinstance(telescope, EarthLocation)
